import logging
import os
import shutil
import random
import signal
import subprocess
import sys
import tempfile
import time
import tarfile
import pickle
from io import BytesIO

import numpy as np

import westpa
from westpa.core.extloader import get_object
from westpa.core.propagators import WESTPropagator
from westpa.core.states import BasisState, InitialState
from westpa.core.segment import Segment
from westpa.core.yamlcfg import check_bool

from westpa.core.trajectory import load_trajectory

log = logging.getLogger(__name__)

# Get a list of user-friendly signal names
SIGNAL_NAMES = {getattr(signal, name): name for name in dir(signal) if name.startswith('SIG') and not name.startswith('SIG_')}


def pcoord_loader(fieldname, pcoord_return_filename, destobj, single_point):
    """Read progress coordinate data into the ``pcoord`` field on ``destobj``.
    An exception will be raised if the data is malformed.  If ``single_point`` is true,
    then only one (N-dimensional) point will be read, otherwise system.pcoord_len points
    will be read.
    """

    system = westpa.rc.get_system_driver()

    assert fieldname == 'pcoord'

    pcoord = np.loadtxt(pcoord_return_filename, dtype=system.pcoord_dtype)

    if single_point:
        expected_shape = (system.pcoord_ndim,)
        if pcoord.ndim == 0:
            pcoord.shape = (1,)
    else:
        expected_shape = (system.pcoord_len, system.pcoord_ndim)
        if pcoord.ndim < 2:
            pcoord.shape = expected_shape
    if pcoord.shape != expected_shape:
        raise ValueError(
            'progress coordinate data has incorrect shape {!r} [expected {!r}] Check pcoord.err or seg_logs for more information.'.format(
                pcoord.shape, expected_shape
            )
        )
    destobj.pcoord = pcoord


def aux_data_loader(fieldname, data_filename, segment, single_point):
    data = np.loadtxt(data_filename)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def npy_data_loader(fieldname, coord_file, segment, single_point):
    log.debug('using npy_data_loader')
    data = np.load(coord_file)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def pickle_data_loader(fieldname, coord_file, segment, single_point):
    log.debug('using pickle_data_loader')
    with open(coord_file, 'rb') as fo:
        data = pickle.load(fo)
    segment.data[fieldname] = data
    if data.nbytes == 0:
        raise ValueError('could not read any data for {}'.format(fieldname))


def trajectory_loader(fieldname, coord_folder, segment, single_point):
    '''Load data from the trajectory return. ``coord_folder`` should be the path to a folder
    containing trajectory files. ``segment`` is the ``Segment`` object that the data is associated with.
    Please see ``load_trajectory`` for more details. ``single_point`` is not used by this loader.'''
    try:
        data = load_trajectory(coord_folder)
        segment.data['iterh5/trajectory'] = data
    except Exception as e:
        log.warning('could not read any data for {}: {}'.format(fieldname, str(e)))


def restart_loader(fieldname, restart_folder, segment, single_point):
    '''Load data from the restart return. The loader will tar all files in ``restart_folder``
    and store it in the per-iteration HDF5 file. ``segment`` is the ``Segment`` object that
    the data is associated with. ``single_point`` is not used by this loader.'''
    try:
        d = BytesIO()
        with tarfile.open(mode='w:gz', fileobj=d) as t:
            t.add(restart_folder, arcname='.')

        segment.data['iterh5/restart'] = d.getvalue() + b'\x01'  # add tail protection
    except Exception as e:
        log.warning('could not read any data for {}: {}'.format(fieldname, str(e)))
    finally:
        d.close()


def restart_writer(path, segment):
    '''Prepare the necessary files from the per-iteration HDF5 file to run ``segment``.'''
    try:
        restart = segment.data.pop('iterh5/restart', None)
        # Making an exception for start states in iteration 1
        if restart is None:
            raise ValueError('restart data is not present')

        d = BytesIO(restart[:-1])  # remove tail protection
        with tarfile.open(fileobj=d, mode='r:gz') as t:
            t.extractall(path=path)
    except ValueError as e:
        log.warning('could not write restart data for {}: {}'.format(str(segment), str(e)))
        d = BytesIO()
        if segment.n_iter == 1:
            log.warning(
                'In iteration 1. Assuming this is a start state and proceeding to skip reading restart from per-iteration HDF5 file for {}'.format(
                    str(segment)
                )
            )
    except Exception as e:
        log.warning('could not write restart data for {}: {}'.format(str(segment), str(e)))
    finally:
        d.close()


def seglog_loader(fieldname, log_file, segment, single_point):
    '''Load data from the log return. The loader will tar all files in ``log_file``
    and store it in the per-iteration HDF5 file. ``segment`` is the ``Segment`` object that
    the data is associated with. ``single_point`` is not used by this loader.'''
    try:
        d = BytesIO()
        with tarfile.open(mode='w:gz', fileobj=d) as t:
            t.add(log_file, arcname='.')

        segment.data['iterh5/log'] = d.getvalue() + b'\x01'  # add tail protection
    except Exception as e:
        log.warning('could not read any data for {}: {}'.format(fieldname, str(e)))
    finally:
        d.close()


class ExecutablePropagator(WESTPropagator):
    ENV_CURRENT_ITER = 'WEST_CURRENT_ITER'

    # Environment variables set during propagation
    ENV_CURRENT_SEG_ID = 'WEST_CURRENT_SEG_ID'
    ENV_CURRENT_SEG_DATA_REF = 'WEST_CURRENT_SEG_DATA_REF'
    ENV_CURRENT_SEG_INITPOINT = 'WEST_CURRENT_SEG_INITPOINT_TYPE'
    ENV_PARENT_SEG_ID = 'WEST_PARENT_ID'
    ENV_PARENT_DATA_REF = 'WEST_PARENT_DATA_REF'

    # Environment variables set during propagation and state generation
    ENV_BSTATE_ID = 'WEST_BSTATE_ID'
    ENV_BSTATE_DATA_REF = 'WEST_BSTATE_DATA_REF'
    ENV_ISTATE_ID = 'WEST_ISTATE_ID'
    ENV_ISTATE_DATA_REF = 'WEST_ISTATE_DATA_REF'

    # Environment variables for progress coordinate calculation
    ENV_STRUCT_DATA_REF = 'WEST_STRUCT_DATA_REF'

    ENV_RAND16 = 'WEST_RAND16'
    ENV_RAND32 = 'WEST_RAND32'
    ENV_RAND64 = 'WEST_RAND64'
    ENV_RAND128 = 'WEST_RAND128'
    ENV_RANDFLOAT = 'WEST_RANDFLOAT'

    def __init__(self, rc=None):
        super().__init__(rc)

        # A mapping of environment variables to template strings which will be
        # added to the environment of all children launched.
        self.addtl_child_environ = dict()

        # A mapping of executable name ('propagator', 'pre_iteration', 'post_iteration') to
        # a dictionary of attributes like 'executable', 'stdout', 'stderr', 'environ', etc.
        self.exe_info = {}
        self.exe_info['propagator'] = {}
        self.exe_info['pre_iteration'] = {}
        self.exe_info['post_iteration'] = {}
        self.exe_info['get_pcoord'] = {}
        self.exe_info['gen_istate'] = {}

        # A mapping of data set name ('pcoord', 'coord', 'com', etc) to a dictionary of
        # attributes like 'loader', 'dtype', etc
        self.data_info = {}
        self.data_info['pcoord'] = {}

        # Validate configuration
        config = self.rc.config

        for key in [
            ('west', 'executable', 'propagator', 'executable'),
            ('west', 'data', 'data_refs', 'segment'),
            ('west', 'data', 'data_refs', 'basis_state'),
            ('west', 'data', 'data_refs', 'initial_state'),
        ]:
            config.require(key)

        self.segment_ref_template = config['west', 'data', 'data_refs', 'segment']
        self.basis_state_ref_template = config['west', 'data', 'data_refs', 'basis_state']
        self.initial_state_ref_template = config['west', 'data', 'data_refs', 'initial_state']
        store_h5 = config.get(['west', 'data', 'data_refs', 'iteration']) is not None

        # Load additional environment variables for all child processes
        self.addtl_child_environ.update({k: str(v) for k, v in (config['west', 'executable', 'environ'] or {}).items()})

        # Load configuration items relating to child processes
        for child_type in ('propagator', 'pre_iteration', 'post_iteration', 'get_pcoord', 'gen_istate', 'subgroup_walkers'):
            child_info = config.get(['west', 'executable', child_type])
            if not child_info:
                continue

            info_prefix = ['west', 'executable', child_type]

            # require executable to be specified if anything is specified at all
            config.require(info_prefix + ['executable'])

            self.exe_info[child_type]['executable'] = child_info['executable']
            self.exe_info[child_type]['stdin'] = child_info.get('stdin', os.devnull)
            self.exe_info[child_type]['stdout'] = child_info.get('stdout', None)
            self.exe_info[child_type]['stderr'] = child_info.get('stderr', None)
            self.exe_info[child_type]['cwd'] = child_info.get('cwd', None)

            if child_type not in ('propagator', 'get_pcoord', 'gen_istate'):
                self.exe_info[child_type]['enabled'] = child_info.get('enabled', True)
            else:
                # for consistency, propagator, get_pcoord, and gen_istate can never be disabled
                self.exe_info[child_type]['enabled'] = True

            # apply environment modifications specific to this executable
            self.exe_info[child_type]['environ'] = {k: str(v) for k, v in (child_info.get('environ') or {}).items()}

        log.debug('exe_info: {!r}'.format(self.exe_info))

        # Load configuration items relating to dataset input
        self.data_info['pcoord'] = {'name': 'pcoord', 'loader': pcoord_loader, 'enabled': True, 'filename': None, 'dir': False}
        self.data_info['trajectory'] = {
            'name': 'trajectory',
            'loader': trajectory_loader,
            'enabled': store_h5,
            'filename': None,
            'dir': True,
        }
        self.data_info['restart'] = {
            'name': 'restart',
            'loader': restart_loader,
            'enabled': store_h5,
            'filename': None,
            'dir': True,
        }
        self.data_info['log'] = {'name': 'seglog', 'loader': seglog_loader, 'enabled': store_h5, 'filename': None, 'dir': False}

        dataset_configs = config.get(['west', 'executable', 'datasets']) or []
        for dsinfo in dataset_configs:
            try:
                dsname = dsinfo['name']
            except KeyError:
                raise ValueError('dataset specifications require a ``name`` field')

            if dsname != 'pcoord':
                check_bool(dsinfo.setdefault('enabled', True))
            else:
                # can never disable pcoord collection
                dsinfo['enabled'] = True

            loader_directive = dsinfo.get('loader')
            if loader_directive:
                if loader_directive == 'default':
                    if dsname not in ['pcoord', 'seglog', 'restart', 'trajectory']:
                        loader = aux_data_loader
                elif loader_directive == 'npy_loader':
                    loader = npy_data_loader
                elif loader_directive == 'pickle_loader':
                    loader = pickle_data_loader
                else:
                    loader = get_object(loader_directive)
                dsinfo['loader'] = loader
            elif dsname not in ['pcoord', 'seglog', 'restart', 'trajectory']:
                loader = aux_data_loader
                dsinfo['loader'] = loader

            self.data_info.setdefault(dsname, {}).update(dsinfo)

        log.debug('data_info: {!r}'.format(self.data_info))

    @staticmethod
    def makepath(template, template_args=None, expanduser=True, expandvars=True, abspath=False, realpath=False):
        template_args = template_args or {}
        path = template.format(**template_args)
        if expandvars:
            path = os.path.expandvars(path)
        if expanduser:
            path = os.path.expanduser(path)
        if realpath:
            path = os.path.realpath(path)
        if abspath:
            path = os.path.abspath(path)
        path = os.path.normpath(path)
        return path

    def random_val_env_vars(self):
        '''Return a set of environment variables containing random seeds. These are returned
        as a dictionary, suitable for use in ``os.environ.update()`` or as the ``env`` argument to
        ``subprocess.Popen()``. Every child process executed by ``exec_child()`` gets these.'''

        return {
            self.ENV_RAND16: str(random.randint(0, 2**16)),
            self.ENV_RAND32: str(random.randint(0, 2**32)),
            self.ENV_RAND64: str(random.randint(0, 2**64)),
            self.ENV_RAND128: str(random.randint(0, 2**128)),
            self.ENV_RANDFLOAT: str(random.random()),
        }

    def exec_child(self, executable, environ=None, stdin=None, stdout=None, stderr=None, cwd=None):
        '''Execute a child process with the environment set from the current environment, the
        values of self.addtl_child_environ, the random numbers returned by self.random_val_env_vars, and
        the given ``environ`` (applied in that order). stdin/stdout/stderr are optionally redirected.

        This function waits on the child process to finish, then returns
        (rc, rusage), where rc is the child's return code and rusage is the resource usage tuple from os.wait4()'''

        all_environ = dict(os.environ)
        all_environ.update(self.addtl_child_environ)
        all_environ.update(self.random_val_env_vars())
        all_environ.update(environ or {})

        stdin = open(stdin, 'rb') if stdin else sys.stdin
        stdout = open(stdout, 'wb') if stdout else sys.stdout
        if stderr == 'stdout':
            stderr = stdout
        else:
            stderr = open(stderr, 'wb') if stderr else sys.stderr

        # close_fds is critical for preventing out-of-file errors
        proc = subprocess.Popen(
            [executable],
            cwd=cwd,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr if stderr != stdout else subprocess.STDOUT,
            close_fds=True,
            env=all_environ,
        )

        # Wait on child and get resource usage
        (_pid, _status, rusage) = os.wait4(proc.pid, 0)
        # Do a subprocess.Popen.wait() to let the Popen instance (and subprocess module) know that
        # we are done with the process, and to get a more friendly return code
        rc = proc.wait()
        return (rc, rusage)

    def exec_child_from_child_info(self, child_info, template_args, environ):
        for (key, value) in child_info.get('environ', {}).items():
            environ[key] = self.makepath(value)
        return self.exec_child(
            executable=self.makepath(child_info['executable'], template_args),
            environ=environ,
            cwd=self.makepath(child_info['cwd'], template_args) if child_info['cwd'] else None,
            stdin=self.makepath(child_info['stdin'], template_args) if child_info['stdin'] else os.devnull,
            stdout=self.makepath(child_info['stdout'], template_args) if child_info['stdout'] else None,
            stderr=self.makepath(child_info['stderr'], template_args) if child_info['stderr'] else None,
        )

    # Functions to create template arguments and environment values for child processes
    def update_args_env_basis_state(self, template_args, environ, basis_state):
        new_template_args = {'basis_state': basis_state}
        new_env = {
            self.ENV_BSTATE_ID: str(basis_state.state_id if basis_state.state_id is not None else -1),
            self.ENV_BSTATE_DATA_REF: self.makepath(self.basis_state_ref_template, new_template_args),
        }
        template_args.update(new_template_args)
        environ.update(new_env)
        return template_args, environ

    def update_args_env_initial_state(self, template_args, environ, initial_state):
        new_template_args = {'initial_state': initial_state}
        new_env = {
            self.ENV_ISTATE_ID: str(initial_state.state_id if initial_state.state_id is not None else -1),
            self.ENV_ISTATE_DATA_REF: self.makepath(self.initial_state_ref_template, new_template_args),
        }

        if initial_state.basis_state is not None:
            basis_state = initial_state.basis_state
        elif initial_state.istate_type == InitialState.ISTATE_TYPE_START:
            basis_state = BasisState(
                label=f"sstate_{initial_state.state_id}", pcoord=initial_state.pcoord, probability=0.0, auxref=""
            )
        else:
            basis_state = self.basis_states[initial_state.basis_state_id]

        self.update_args_env_basis_state(new_template_args, new_env, basis_state)

        template_args.update(new_template_args)
        environ.update(new_env)
        return template_args, environ

    def update_args_env_iter(self, template_args, environ, n_iter):
        environ[self.ENV_CURRENT_ITER] = str(n_iter if n_iter is not None else -1)
        template_args['n_iter'] = int(n_iter)
        return template_args, n_iter

    def update_args_env_segment(self, template_args, environ, segment):
        template_args['segment'] = segment

        environ[self.ENV_CURRENT_SEG_INITPOINT] = Segment.initpoint_type_names[segment.initpoint_type]

        if segment.initpoint_type == Segment.SEG_INITPOINT_CONTINUES:
            # Could use actual parent object here if the work manager cared to pass that much data
            # to us (we'd need at least the subset of parents for all segments sent in the call to propagate)
            # that may make a good west.cfg option for future crazy extensibility, but for now,
            # just populate the bare minimum
            parent = Segment(n_iter=segment.n_iter - 1, seg_id=segment.parent_id)
            parent_template_args = dict(template_args)
            parent_template_args['segment'] = parent

            environ[self.ENV_PARENT_SEG_ID] = str(segment.parent_id if segment.parent_id is not None else -1)
            environ[self.ENV_PARENT_DATA_REF] = self.makepath(self.segment_ref_template, parent_template_args)
        elif segment.initpoint_type == Segment.SEG_INITPOINT_NEWTRAJ:
            # This segment is initiated from a basis state; WEST_PARENT_SEG_ID and WEST_PARENT_DATA_REF are
            # set to the basis state ID and data ref
            initial_state = self.initial_states[segment.initial_state_id]

            if initial_state.istate_type == InitialState.ISTATE_TYPE_START:

                basis_state = BasisState(
                    label=f"sstate_{initial_state.state_id}", pcoord=initial_state.pcoord, probability=0.0, auxref=""
                )

            else:
                basis_state = self.basis_states[initial_state.basis_state_id]

            if self.ENV_BSTATE_ID not in environ:
                self.update_args_env_basis_state(template_args, environ, basis_state)
            if self.ENV_ISTATE_ID not in environ:
                self.update_args_env_initial_state(template_args, environ, initial_state)

            assert initial_state.istate_type in (
                InitialState.ISTATE_TYPE_BASIS,
                InitialState.ISTATE_TYPE_GENERATED,
                InitialState.ISTATE_TYPE_START,
            )
            if initial_state.istate_type == InitialState.ISTATE_TYPE_BASIS:
                environ[self.ENV_PARENT_DATA_REF] = environ[self.ENV_BSTATE_DATA_REF]

            elif initial_state.istate_type == InitialState.ISTATE_TYPE_START:

                # This points to the start-state PDB
                environ[self.ENV_PARENT_DATA_REF] = environ[self.ENV_BSTATE_DATA_REF] + '/' + initial_state.basis_auxref
            else:  # initial_state.type == InitialState.ISTATE_TYPE_GENERATED
                environ[self.ENV_PARENT_DATA_REF] = environ[self.ENV_ISTATE_DATA_REF]

        environ[self.ENV_CURRENT_SEG_ID] = str(segment.seg_id if segment.seg_id is not None else -1)
        environ[self.ENV_CURRENT_SEG_DATA_REF] = self.makepath(self.segment_ref_template, template_args)
        return template_args, environ

    def template_args_for_segment(self, segment):
        template_args, environ = {}, {}
        self.update_args_env_iter(template_args, environ, segment.n_iter)
        self.update_args_env_segment(template_args, environ, segment)
        return template_args

    def exec_for_segment(self, child_info, segment, addtl_env=None):
        '''Execute a child process with environment and template expansion from the given
        segment.'''
        template_args, environ = {}, {}
        self.update_args_env_iter(template_args, environ, segment.n_iter)
        self.update_args_env_segment(template_args, environ, segment)
        environ.update(addtl_env or {})
        self.prepare_file_system(segment, environ)
        child_info['cwd'] = environ[self.ENV_CURRENT_SEG_DATA_REF]
        return self.exec_child_from_child_info(child_info, template_args, environ)

    def exec_for_iteration(self, child_info, n_iter, addtl_env=None):
        '''Execute a child process with environment and template expansion from the given
        iteration number.'''
        template_args, environ = {}, {}
        self.update_args_env_iter(template_args, environ, n_iter)
        environ.update(addtl_env or {})
        return self.exec_child_from_child_info(child_info, template_args, environ)

    def exec_for_basis_state(self, child_info, basis_state, addtl_env=None):
        '''Execute a child process with environment and template expansion from the
        given basis state'''
        template_args, environ = {}, {}
        self.update_args_env_basis_state(template_args, environ, basis_state)
        environ.update(addtl_env or {})
        return self.exec_child_from_child_info(child_info, template_args, environ)

    def exec_for_initial_state(self, child_info, initial_state, addtl_env=None):
        '''Execute a child process with environment and template expansion from the given
        initial state.'''
        template_args, environ = {}, {}
        self.update_args_env_initial_state(template_args, environ, initial_state)
        environ.update(addtl_env or {})
        return self.exec_child_from_child_info(child_info, template_args, environ)

    def prepare_file_system(self, segment, environ):
        try:
            # If the filesystem is properly clean.
            os.makedirs(environ[self.ENV_CURRENT_SEG_DATA_REF])
        except Exception:
            # If the filesystem is NOT properly clean.
            shutil.rmtree(environ[self.ENV_CURRENT_SEG_DATA_REF])
            os.makedirs(environ[self.ENV_CURRENT_SEG_DATA_REF])
        if self.data_info['restart']['enabled']:
            restart_writer(environ[self.ENV_CURRENT_SEG_DATA_REF], segment=segment)

    def setup_dataset_return(self, segment=None, subset_keys=None):
        '''Set up temporary files and environment variables that point to them for segment
        runners to return data. ``segment`` is the ``Segment`` object that the return data
        is associated with. ``subset_keys`` specifies the names of a subset of data to be
        returned.'''
        if subset_keys is None:
            subset_keys = self.data_info.keys()

        addtl_env = {}
        return_files = {}
        del_return_files = {}

        for dataset in self.data_info:
            if dataset not in subset_keys:
                continue

            if not self.data_info[dataset].get('enabled', False):
                continue

            return_template = self.data_info[dataset].get('filename')
            if return_template:
                if segment is None:
                    raise ValueError('segment needs to be provided for dataset return')
                return_files[dataset] = self.makepath(return_template, self.template_args_for_segment(segment))
                del_return_files[dataset] = False
            else:
                isdir = self.data_info[dataset].get('dir', False)
                if isdir:
                    rfname = tempfile.mkdtemp()
                else:
                    (fd, rfname) = tempfile.mkstemp()
                    os.close(fd)
                return_files[dataset] = rfname
                del_return_files[dataset] = True

            addtl_env['WEST_{}_RETURN'.format(dataset.upper())] = return_files[dataset]

        return addtl_env, return_files, del_return_files

    def retrieve_dataset_return(self, segment, return_files, del_return_files, single_point):
        '''Retrieve returned data from the temporary locations directed by the environment variables.
        ``segment`` is the ``Segment`` object that the return data is associated with. ``return_files``
        is a ``dict`` where the keys are the dataset names and the values are the paths to the temporarily
        files that contain the returned data. ``del_return_files`` is a ``dict`` where the keys are the
        names of datasets to be deleted (if the corresponding value is set to ``True``) once the data is
        retrieved.'''
        for dataset in self.data_info:
            if dataset not in return_files:
                continue

            # pcoord is always enabled (see __init__)
            if not self.data_info[dataset].get('enabled', False):
                continue

            filename = return_files[dataset]
            loader = self.data_info[dataset]['loader']
            try:
                loader(dataset, filename, segment, single_point=single_point)
            except Exception as e:
                log.error('could not read {} from {!r}: {!r}'.format(dataset, filename, e))
                segment.status = Segment.SEG_STATUS_FAILED
                break
            else:
                if del_return_files.get(dataset, False):
                    try:
                        if os.path.isfile(filename):
                            os.unlink(filename)
                        else:
                            shutil.rmtree(filename)
                    except Exception as e:
                        log.warning('could not delete {} file {!r}: {!r}'.format(dataset, filename, e))
                    else:
                        log.debug('deleted {} file {!r}'.format(dataset, filename))

    # Specific functions required by the WEST framework
    def get_pcoord(self, state):
        '''Get the progress coordinate of the given basis or initial state.'''

        template_args, environ = {}, {}

        if isinstance(state, BasisState):
            execfn = self.exec_for_basis_state
            self.update_args_env_basis_state(template_args, environ, state)
            struct_ref = environ[self.ENV_BSTATE_DATA_REF]
        elif isinstance(state, InitialState):
            execfn = self.exec_for_initial_state
            self.update_args_env_initial_state(template_args, environ, state)
            struct_ref = environ[self.ENV_ISTATE_DATA_REF]
        else:
            raise TypeError('state must be a BasisState or InitialState')

        child_info = self.exe_info.get('get_pcoord')
        addtl_env, return_files, del_return_files = self.setup_dataset_return(
            subset_keys=['pcoord', 'trajectory', 'restart', 'log']
        )
        addtl_env[self.ENV_STRUCT_DATA_REF] = struct_ref

        rc, rusage = execfn(child_info, state, addtl_env)
        if rc != 0:
            log.error('get_pcoord executable {!r} returned {}'.format(child_info['executable'], rc))

        self.retrieve_dataset_return(state, return_files, del_return_files, True)

    def gen_istate(self, basis_state, initial_state):
        '''Generate a new initial state from the given basis state.'''
        child_info = self.exe_info.get('gen_istate')
        rc, rusage = self.exec_for_initial_state(child_info, initial_state)
        if rc != 0:
            log.error('gen_istate executable {!r} returned {}'.format(child_info['executable'], rc))
            initial_state.istate_status = InitialState.ISTATE_STATUS_FAILED
            return

        # Determine and load the progress coordinate value for this state
        try:
            self.get_pcoord(initial_state)
        except Exception:
            log.exception('could not get progress coordinate for initial state {!r}'.format(initial_state))
            initial_state.istate_status = InitialState.ISTATE_STATUS_FAILED
            raise
        else:
            initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED

    def prepare_iteration(self, n_iter, segments):
        child_info = self.exe_info.get('pre_iteration')
        if child_info and child_info['enabled']:
            try:
                rc, rusage = self.exec_for_iteration(child_info, n_iter)
            except OSError as e:
                log.warning('could not execute pre-iteration program {!r}: {}'.format(child_info['executable'], e))
            else:
                if rc != 0:
                    log.warning('pre-iteration executable {!r} returned {}'.format(child_info['executable'], rc))

    def finalize_iteration(self, n_iter, segments):
        child_info = self.exe_info.get('post_iteration')
        if child_info and child_info['enabled']:
            try:
                rc, rusage = self.exec_for_iteration(child_info, n_iter)
            except OSError as e:
                log.warning('could not execute post-iteration program {!r}: {}'.format(child_info['executable'], e))
            else:
                if rc != 0:
                    log.warning('post-iteration executable {!r} returned {}'.format(child_info['executable'], rc))

    def propagate(self, segments):
        child_info = self.exe_info['propagator']

        for segment in segments:
            starttime = time.time()

            addtl_env, return_files, del_return_files = self.setup_dataset_return(segment)

            # Spawn propagator and wait for its completion
            rc, rusage = self.exec_for_segment(child_info, segment, addtl_env)

            if rc == 0:
                segment.status = Segment.SEG_STATUS_COMPLETE
            elif rc < 0:
                log.error('child process for segment %d exited on signal %d (%s)' % (segment.seg_id, -rc, SIGNAL_NAMES[-rc]))
                segment.status = Segment.SEG_STATUS_FAILED
                continue
            else:
                log.error('child process for segment %d exited with code %d' % (segment.seg_id, rc))
                segment.status = Segment.SEG_STATUS_FAILED
                continue

            # Extract data and store on segment for recording in the master thread/process/node
            self.retrieve_dataset_return(segment, return_files, del_return_files, False)

            if segment.status == Segment.SEG_STATUS_FAILED:
                continue

            # Record timing info
            segment.walltime = time.time() - starttime
            segment.cputime = rusage.ru_utime
        return segments



from westpa.tools import WESTMasterCommand, WESTParallelTool
from warnings import warn

from westpa.cli.tools.w_reweight import RWMatrix

# Just a shim to make sure everything works and is backwards compatible.
# We're making sure it has the appropriate functions so that it can be called
# as a regular tool, and not a subcommand.


class PAMatrix(RWMatrix):
    subcommand = 'init'
    help_text = 'averages and CIs for path-tracing kinetics analysis'
    default_output_file = 'flux_matrices.h5'
    # This isn't strictly necessary, but for the moment, here it is.
    # We really need to modify the underlying class so that we don't pull this sort of stuff if it isn't necessary.
    # That'll take some case handling, which is fine.
    # default_kinetics_file = 'assign.h5'


class WReweight(WESTMasterCommand, WESTParallelTool):
    prog = 'w_postanalysis_matrix'
    subcommands = [PAMatrix]
    subparsers_title = 'calculate state-to-state kinetics by tracing trajectories'
    description = '''\
Generate a colored transition matrix from a WE assignment file. The subsequent
analysis requires that the assignments are calculated using only the initial and
final time points of each trajectory segment. This may require downsampling the
h5file generated by a WE simulation. In the future w_assign may be enhanced to optionally
generate the necessary assignment file from a h5file with intermediate time points.
Additionally, this analysis is currently only valid on simulations performed under
either equilibrium or steady-state conditions without recycling target states.

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "reweight.h5") contains the
following datasets:

  ``/bin_populations`` [window, bin]
     The reweighted populations of each bin based on windows. Bins contain
     one color each, so to recover the original un-colored spatial bins,
     one must sum over all states.

  ``/iterations`` [iteration]
    *(Structured -- see below)*  Sparse matrix data from each
    iteration.  They are reconstructed and averaged within the
    w_reweight {kinetics/probs} routines so that observables may
    be calculated.  Each group contains 4 vectors of data:

      flux
        *(Floating-point)* The weight of a series of flux events
      cols
        *(Integer)* The bin from which a flux event began.
      cols
        *(Integer)* The bin into which the walker fluxed.
      obs
        *(Integer)* How many flux events were observed during this
        iteration.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''


def entry_point():
    warn('{} is being deprecated.  Please use w_reweight instead.'.format(WReweight.prog))
    # If we're not really supporting subcommands...
    import sys

    try:
        if sys.argv[1] != 'init':
            sys.argv.insert(1, 'init')
    except Exception:
        sys.argv.insert(1, 'init')
    WReweight().main()


if __name__ == '__main__':
    entry_point()



import logging

import numpy as np

from westpa.oldtools.aframe import AnalysisMixin

log = logging.getLogger(__name__)

try:
    import matplotlib
except ImportError:
    matplotlib = None
    pyplot = None
    cm_hovmol = None
    cm_hovmol_r = None
else:
    try:
        matplotlib.use('PDF')
        from matplotlib import pyplot
    except Exception as e:
        log.info('could not select matplotlib PDF backend: {}'.format(e))
        matplotlib = None
        pyplot = None

    cmap_data = np.array(
        [
            (124, 0, 24),
            (211, 0, 32),
            (244, 184, 0),
            (245, 235, 0),
            (129, 183, 2),
            (32, 128, 38),
            (21, 27, 87),
            (36, 62, 137),
            (178, 220, 245),
            (255, 255, 255),
        ],
        np.float32,
    )
    cmap_data /= 255.0
    cm_hovmol = matplotlib.colors.LinearSegmentedColormap.from_list('hovmol', cmap_data)
    cm_hovmol_r = matplotlib.colors.LinearSegmentedColormap.from_list('hovmol_r', np.flipud(cmap_data))

    _pdr_data = {
        'red': [
            (0.0, 1.0, 1.0),
            (0.25, 1.0, 1.0),
            (0.50, 1.0, 1.0),
            (0.75, 0.0, 0.0),
            (0.90, 0.50, 0.50),
            (0.95, 0.0, 0.0),
            (1.0, 1.0, 1.0),
        ],
        'green': [
            (0.0, 1.0, 1.0),
            (0.25, 0.0, 0.0),
            (0.50, 1.0, 1.0),
            (0.75, 1.0, 1.0),
            (0.90, 0.86, 0.86),
            (0.95, 0.0, 0.0),
            (1.0, 1.0, 1.0),
        ],
        'blue': [(0.0, 1.0, 1.0), (0.25, 0.0, 0.0), (0.75, 0.0, 0.0), (0.90, 1.0, 1.0), (0.95, 1.0, 1.0), (1.0, 1.0, 1.0)],
    }

    cm_pdr = matplotlib.colors.LinearSegmentedColormap('pdr', _pdr_data, 2048)
    cm_pdr_r = cm_pdr.reversed()

    matplotlib.cm.register_cmap('pdr', cm_pdr)
    matplotlib.cm.register_cmap('pdr_r', cm_pdr_r)
    matplotlib.cm.register_cmap('hovmol', cm_hovmol)
    matplotlib.cm.register_cmap('hovmol_r', cm_hovmol_r)

    del cmap_data


class PlottingMixin(AnalysisMixin):
    def __init__(self):
        global matplotlib, pyplot

        super().__init__()

        self.matplotlib_avail = matplotlib is not None and pyplot is not None

    def require_matplotlib(self):
        global matplotlib
        if not self.matplotlib_avail:
            raise RuntimeError('matplotlib is not available')
        else:
            return matplotlib



import numpy as np
import numpy.ma as ma

from . import UncertMath
from . import BinCluster


def probAdjustEquil(binProb, rates, uncert, threshold=0.0, fullCalcClust=False, fullCalcBins=False):
    """This function adjusts bin pops in binProb using rates and uncert matrices
    fullCalcBins --> True for weighted avg, False for simple calc
    fullCalcClust --> True for weighted avg, False for simple calc
    threshold --> minimum weight (relative to max) for another value to be averaged
            only matters if fullCalcBins == True (or later perhaps if fullCalcClust == True)
    """

    # Check that rate matrix is square
    Ni, Nj = rates.shape
    if Ni != Nj:
        print('\nWARNING: Not a square matrix!\n')

    zi = np.where(binProb == 0.0)[0]  # indices of bins with zero probability

    rates_uncert = UncertMath.UncertContainer(rates, rates - uncert, rates + uncert)

    # STEP 1a: Create matrix of ratios of probabilities based on DIRECT estimates
    # that is, ij element is p_i / p_j = k_ji / k_ij

    ratios_direct = rates_uncert.transpose() / rates_uncert

    # STEP 1b: Create averaged matrix of ratios of probabilities based on both direct and indirect estimates
    # Indirect means '3rd bin' estimates: p_i / p_j = ( k_ki / k_ik ) ( k_jk / k_kj )
    # Turns out this is not helpful, so generally set fullCalcBins = 0
    if fullCalcBins:
        # Calculate indirect ratios using Einstein Summation convention where
        # ratios_indirect_kij  = ( k_ki / k_ik ) ( k_jk / k_kj ) = ratios_direct_ik * ratios_direct_kj
        ri_vals = np.einsum('ik,kj->kij', ratios_direct.vals, ratios_direct.vals)
        ri_min = np.einsum('ik,kj->kij', ratios_direct.dmin, ratios_direct.dmin)
        ri_max = np.einsum('ik,kj->kij', ratios_direct.dmax, ratios_direct.dmax)
        ratios_indirect = UncertMath.UncertContainer(ri_vals, ri_min, ri_max, mask=ratios_direct.vals.mask)

        # Threshold indirect ratios
        ti = ratios_indirect.wt < ratios_direct * threshold
        ratios_indirect.mask = ti
        ratios_indirect.update_mask()

        ratios_indirect.concatenate(ratios_direct, axis=0)
        ratios_average = ratios_indirect.weighted_average(axis=0)

    else:
        ratios_average = ratios_direct.weighted_average(axis=0, expaxis=0)

    # STEP 2: Form clusters

    # STEP 2a: Sort probability ratios based on uncertainty
    # Sort uncertainties of ratios_average subject to the convention that p_i < p_j

    i, j = np.triu_indices(Ni, 1)  # indices of ij pairs where i != j

    # Remove pairs that include a bin that has zero probability
    nzi = (binProb[i] != 0.0) & (binProb[j] != 0.0)
    i = i[nzi]
    j = j[nzi]

    vals = ma.vstack((ratios_average.vals[i, j], ratios_average.vals[j, i]))
    ias = ma.argsort(vals, axis=0, fill_value=np.inf)

    ordered_ind = np.vstack((i, j))
    flip_ind = np.nonzero(ias[0, :])  # Find pairs in which to select ji rather than ij
    ordered_ind[:, flip_ind[0]] = ordered_ind[:, flip_ind[0]][::-1]

    iind = ordered_ind[0, :]
    jind = ordered_ind[1, :]
    uncertij = ratios_average.uncert[iind, jind]  # Get the uncert for ij pairs

    count = uncertij.count()  # Count of the unmasked uncertainties
    ias = ma.argsort(uncertij, fill_value=np.inf)  # Get the indices that would sort uncertij
    iind = iind[ias[:count]]  # Sort the indices excluding masked/undefined values
    jind = jind[ias[:count]]

    # STEP 2b: Create ClusterList object and cluster bins
    clusters = BinCluster.ClusterList(ratios_average, Ni)

    if fullCalcClust:
        clusters.join((iind, jind))
    else:
        clusters.join_simple((iind, jind))

    total_prob = 0.0  # total probability in all clusters
    for cid in clusters.cluster_contents:
        binlist = list(clusters.cluster_contents[cid])
        if len(binlist):
            prob_cluster = binProb[binlist].sum()
            total_prob += prob_cluster

            binProb[binlist] = prob_cluster * clusters.bin_data[binlist].vals

    binProb[zi] = 0.0  # re-zero bins that previously had zero prob
    # for bi,p in enumerate(binProb):
    #    print('bin: {} -- {}'.format(bi,p))
    print('.........Total Probability: {}'.format(binProb.sum()))



import logging

import numpy as np
from westpa.core.we_driver import WEDriver

log = logging.getLogger(__name__)


class MABDriver(WEDriver):
    def assign(self, segments, initializing=False):
        '''Assign segments to initial and final bins, and update the (internal) lists of used and available
        initial states. This function is adapted to the MAB scheme, so that the inital and final segments are
        sent to the bin mapper at the same time, otherwise the inital and final bin boundaries can be inconsistent.'''

        log.debug("MABDriver in use.")
        # collect initial and final coordinates into one place
        n_segments = len(segments)
        all_pcoords = np.empty((n_segments * 2, self.system.pcoord_ndim + 2), dtype=self.system.pcoord_dtype)

        for iseg, segment in enumerate(segments):
            all_pcoords[iseg] = np.append(segment.pcoord[0, :], [segment.weight, 0.0])
            all_pcoords[n_segments + iseg] = np.append(segment.pcoord[-1, :], [segment.weight, 1.0])

        # assign based on initial and final progress coordinates
        assignments = self.bin_mapper.assign(all_pcoords)
        initial_assignments = assignments[:n_segments]
        if initializing:
            final_assignments = initial_assignments
        else:
            final_assignments = assignments[n_segments:]

        initial_binning = self.initial_binning
        final_binning = self.final_binning
        flux_matrix = self.flux_matrix
        transition_matrix = self.transition_matrix
        for (segment, iidx, fidx) in zip(segments, initial_assignments, final_assignments):
            initial_binning[iidx].add(segment)
            final_binning[fidx].add(segment)
            flux_matrix[iidx, fidx] += segment.weight
            transition_matrix[iidx, fidx] += 1

        n_recycled_total = self.n_recycled_segs
        n_new_states = n_recycled_total - len(self.avail_initial_states)

        log.debug(
            '{} walkers scheduled for recycling, {} initial states available'.format(
                n_recycled_total, len(self.avail_initial_states)
            )
        )

        if n_new_states > 0:
            return n_new_states
        else:
            return 0



'''
YAML-based configuration files for WESTPA
'''

import os
import warnings

import yaml
import numpy as np

try:
    from yaml import CLoader as YLoader
except ImportError:
    # fall back on Python implementation
    from yaml import Loader as YLoader

from . import extloader
from .binning import NopMapper

# Only needed for temporary class
import westpa

NotProvided = object()


class ConfigValueWarning(UserWarning):
    pass


def warn_dubious_config_entry(entry, value, expected_type=None, category=ConfigValueWarning, stacklevel=1):
    if expected_type:
        warnings.warn(
            'dubious configuration entry {}: {} (expected type {})'.format(entry, value, expected_type), category, stacklevel + 1
        )
    else:
        warnings.warn('dubious configuration entry {}: {}'.format(entry, value), category, stacklevel + 1)


def check_bool(value, action='warn'):
    '''Check that the given ``value`` is boolean in type. If not, either
    raise a warning (if ``action=='warn'``) or an exception (``action=='raise'``).
    '''
    if action not in ('warn', 'raise'):
        raise ValueError('invalid action {!r}'.format(action))

    if not isinstance(value, bool):
        if action == 'warn':
            warnings.warn(
                'dubious boolean value {!r}, will be treated as {!r}'.format(value, bool(value)),
                category=ConfigValueWarning,
                stacklevel=2,
            )
        elif action == 'raise':
            raise ValueError('dubious boolean value {!r}, would be treated as {!r}'.format(value, bool(value)))
    else:
        return value


class ConfigItemMissing(KeyError):
    def __init__(self, key, message=None):
        self.key = key
        if message is None:
            message = 'configuration item missing: {!r}'.format(key)
        super().__init__(message)


class ConfigItemTypeError(TypeError):
    def __init__(self, key, expected_type, message=None):
        self.key = key
        self.expected_type = expected_type
        if message is None:
            message = 'configuration item {!r} must have type {!r}'.format(key, expected_type)
        super().__init__(message)


class ConfigValueError(ValueError):
    def __init__(self, key, value, message=None):
        self.key = key
        self.value = value
        if message is None:
            message = 'bad value {!r} for configuration item {!r}'.format(key, value)
        super().__init__(message)


class YAMLConfig:
    preload_config_files = ['/etc/westpa/westrc', os.path.expanduser('~/.westrc')]

    def __init__(self):
        self._data = {}

        for source in self.preload_config_files:
            self.update_from_file(source, required=False)

    def __repr__(self):
        return repr(self._data)

    def update_from_file(self, file, required=True):
        if isinstance(file, str):
            try:
                file = open(file, 'rt')
            except IOError:
                if required:
                    raise
                else:
                    return

        self._data.update(yaml.load(file, Loader=YLoader))
        file.close()

    def _normalize_key(self, key):
        if isinstance(key, str):
            key = (key,)
        else:
            try:
                key = tuple(key)
            except TypeError:
                key = (key,)
        return key

    def _resolve_object_chain(self, key, last=None):
        if last is None:
            last = len(key)
        objects = [self._data[key[0]]]
        for subkey in key[1:last]:
            objects.append(objects[-1][subkey])
        return objects

    def __getitem__(self, key):
        key = self._normalize_key(key)
        return self._resolve_object_chain(key)[-1]

    def __setitem__(self, key, value):
        key = self._normalize_key(key)

        try:
            objchain = self._resolve_object_chain(key, -1)
        except KeyError:
            # creation of a new (possibly nested) entry
            val = self._data
            for keypart in key[:-1]:
                try:
                    val = val[keypart]
                except KeyError:
                    val[keypart] = {}
            try:
                val = val[key[-1]]
            except KeyError:
                val[key[-1]] = value
        else:
            objchain[-1][key[-1]] = value

    def __delitem__(self, key):
        key = self._normalize_key(key)
        objchain = self._resolve_object_chain(key, -1)
        del objchain[-1][key[-1]]

    def __contains__(self, key):
        try:
            self[key]
        except (KeyError, TypeError):
            return False
        else:
            return True

    def require(self, key, type_=None):
        '''Ensure that a configuration item with the given ``key`` is present. If
        the optional ``type_`` is given, additionally require that the item has that
        type.'''

        try:
            item = self[key]
        except KeyError:
            raise ConfigItemMissing(key)

        if type_ is not None:
            if not isinstance(item, type_):
                raise ConfigItemTypeError(item, type_)
        return item

    def require_type_if_present(self, key, type_):
        '''Ensure that the configuration item with the given ``key`` has the
        given type.'''

        try:
            item = self[key]
        except KeyError:
            return
        else:
            if not isinstance(item, type_):
                raise ConfigItemTypeError(item, type_)

    def coerce_type_if_present(self, key, type_):
        try:
            item = self[key]
        except KeyError:
            return
        else:
            if type_ is bool and not isinstance(item, bool):
                warn_dubious_config_entry(key, item, bool)
            self[key] = type_(item)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def get_typed(self, key, type_, default=NotProvided):
        try:
            item = self[key]
        except KeyError as ke:
            if default is not NotProvided:
                item = default
            else:
                raise ke

        # Warn about possibly bad boolean
        if type_ is bool and not isinstance(item, bool):
            warn_dubious_config_entry(key, item, bool)

        return type_(item)

    def get_path(self, key, default=NotProvided, expandvars=True, expanduser=True, realpath=True, abspath=True):
        try:
            path = self[key]
        except KeyError as ke:
            if default is not NotProvided:
                path = default
            else:
                raise ke

        if expandvars:
            path = os.path.expandvars(path)
        if expanduser:
            path = os.path.expanduser(path)
        if realpath:
            path = os.path.realpath(path)
        if abspath:
            path = os.path.abspath(path)

        return path

    def get_pathlist(self, key, default=NotProvided, sep=os.pathsep, expandvars=True, expanduser=True, realpath=True, abspath=True):
        try:
            paths = self[key]
        except KeyError as ke:
            if default is not NotProvided:
                paths = default
            else:
                raise ke

        try:
            items = paths.split(sep)
        except AttributeError:
            # Default must have been something we can't process, like a list or None
            # Just pass it through, since enforcing a restriction on what kind of
            # default is passed is probably more counterproductive than any poor programming
            # practice it encourages.
            return paths

        if expandvars:
            items = list(map(os.path.expandvars, items))
        if expanduser:
            items = list(map(os.path.expanduser, items))
        if realpath:
            items = list(map(os.path.realpath, items))
        if abspath:
            items = list(map(os.path.abspath, items))

        return items

    def get_python_object(self, key, default=NotProvided, path=None):
        try:
            qualname = self[key]
        except KeyError as ke:
            if default is not NotProvided:
                return default
            else:
                raise ke

        return extloader.get_object(qualname, path)

    def get_choice(self, key, choices, default=NotProvided, value_transform=None):
        try:
            value = self[key]
        except KeyError:
            if default is not NotProvided:
                value = default
            else:
                raise

        choices = set(choices)
        if value_transform:
            value = value_transform(value)
        if value not in choices:
            raise ConfigValueError(
                key,
                value,
                message='bad value {!r} for configuration item {!r} (valid choices: {!r})'.format(
                    value, key, tuple(sorted(choices))
                ),
            )
        return value


# Temporary class here
class YAMLSystem:
    '''A description of the system being simulated, including the dimensionality and
    data type of the progress coordinate, the number of progress coordinate entries
    expected from each segment, and binning. To construct a simulation, the user must
    subclass WESTSystem and set several instance variables.

    At a minimum, the user must subclass ``WESTSystem`` and override
    :method:`initialize` to set the data type and dimensionality of progress
    coordinate data and define a bin mapper.

    :ivar pcoord_ndim:    The number of dimensions in the progress coordinate.
                          Defaults to 1 (i.e. a one-dimensional progress
                          coordinate).
    :ivar pcoord_dtype:   The data type of the progress coordinate, which must be
                          callable (e.g. ``np.float32`` and ``long`` will work,
                          but ``'<f4'`` and ``'<i8'`` will not).  Defaults to
                          ``np.float64``.
    :ivar pcoord_len:     The length of the progress coordinate time series
                          generated by each segment, including *both* the initial
                          and final values.  Defaults to 2 (i.e. only the initial
                          and final progress coordinate values for a segment are
                          returned from propagation).
    :ivar bin_mapper:     A bin mapper describing the progress coordinate space.
    :ivar bin_target_counts: A vector of target counts, one per bin.
    '''

    def __init__(self, rc=None):
        self.rc = rc or westpa.rc

        # Number of dimentions in progress coordinate data
        self.pcoord_ndim = 1

        # Length of progress coordinate data for each segment
        self.pcoord_len = 2

        # Data type of progress coordinate
        self.pcoord_dtype = np.float32

        # Mapper
        self.bin_mapper = NopMapper()
        # self.bin_mapper = None
        self._bin_target_counts = None

        self.bin_target_counts = [1]

    @property
    def bin_target_counts(self):
        return self._bin_target_counts

    @bin_target_counts.setter
    def bin_target_counts(self, target_counts):
        maxcount = max(target_counts)
        self._bin_target_counts = np.array(target_counts, dtype=np.min_scalar_type(maxcount))

    def initialize(self):
        '''Prepare this system object for use in simulation or analysis,
        creating a bin space, setting replicas per bin, and so on. This
        function is called whenever a WEST tool creates an instance of the
        system driver.
        '''
        pass

    def prepare_run(self):
        '''Prepare this system for use in a simulation run. Called by w_run in
        all worker processes.'''
        pass

    def finalize_run(self):
        '''A hook for system-specific processing for the end of a simulation run
        (as defined by such things as maximum wallclock time, rather than perhaps
        more scientifically-significant definitions of "the end of a simulation run")'''
        pass

    def new_pcoord_array(self, pcoord_len=None):
        '''Return an appropriately-sized and -typed pcoord array for a timepoint, segment,
        or number of segments. If ``pcoord_len`` is not specified (or None), then
        a length appropriate for a segment is returned.'''

        if pcoord_len is None:
            pcoord_len = self.pcoord_len
        return np.zeros((pcoord_len, self.pcoord_ndim), self.pcoord_dtype)

    def new_region_set(self):
        raise NotImplementedError('This method has been removed.')



import json
import logging
import multiprocessing
import socket

from .core import ZMQCore, Message, Task, Result, ZMQWorkerMissing, ZMQWMEnvironmentError, IsNode
from .core import PassiveMultiTimer
from .core import randport
from .worker import ZMQWorker
from .node import ZMQNode

import westpa.work_managers as work_managers
from westpa.work_managers import WorkManager, WMFuture

import zmq

from collections import deque

log = logging.getLogger(__name__)


class ZMQWorkManager(ZMQCore, WorkManager, IsNode):
    @classmethod
    def add_wm_args(cls, parser, wmenv=None):
        if wmenv is None:
            wmenv = work_managers.environment.default_env

        wm_group = parser.add_argument_group('options for ZeroMQ ("zmq") work manager (master or node)')

        wm_group.add_argument(
            wmenv.arg_flag('zmq_mode'),
            metavar='MODE',
            choices=('master', 'node', 'server', 'client'),
            help='Operate as a master (server) or a node (workers/client). '
            + '"server" is a deprecated synonym for "master" and "client" is a '
            + 'deprecated synonym for "node".',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_comm_mode'),
            metavar='COMM_MODE',
            choices=('ipc', 'tcp'),
            help='Use the given communication mode -- TCP or IPC (Unix-domain) -- sockets '
            + 'for communication within a node. IPC (the default) may be more '
            + 'efficient but is not available on (exceptionally rare) systems '
            + 'without node-local storage (e.g. /tmp); on such systems, TCP may be used instead.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_write_host_info'),
            metavar='INFO_FILE',
            help='Store hostname and port information needed to connect to this instance '
            + 'in INFO_FILE. This allows the master and nodes assisting in '
            + 'coordinating the communication of other nodes to choose ports '
            + 'randomly. Downstream nodes read this file with '
            + wmenv.arg_flag('zmq_read_host_info')
            + ' and know where how to connect.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_read_host_info'),
            metavar='INFO_FILE',
            help='Read hostname and port information needed to connect to the master '
            + '(or other coordinating node) from INFO_FILE. '
            + 'This allows the master and nodes assisting in '
            + 'coordinating the communication of other nodes to choose ports '
            + 'randomly, writing that information with '
            + wmenv.arg_flag('zmq_write_host_info')
            + ' for this instance to read.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_upstream_rr_endpoint'),
            metavar='ENDPOINT',
            help='ZeroMQ endpoint to which to send request/response (task and result) ' + 'traffic toward the master.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_upstream_ann_endpoint'),
            metavar='ENDPOINT',
            help='ZeroMQ endpoint on which to receive announcement '
            + '(heartbeat and shutdown notification) traffic from the master.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_downstream_rr_endpoint'),
            metavar='ENDPOINT',
            help='ZeroMQ endpoint on which to listen for request/response ' + '(task and result) traffic from subsidiary workers.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_downstream_ann_endpoint'),
            metavar='ENDPOINT',
            help='ZeroMQ endpoint on which to send announcement ' + '(heartbeat and shutdown notification) traffic toward workers.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_master_heartbeat'),
            metavar='MASTER_HEARTBEAT',
            type=float,
            help='Every MASTER_HEARTBEAT seconds, the master announces its presence ' + 'to workers.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_worker_heartbeat'),
            metavar='WORKER_HEARTBEAT',
            type=float,
            help='Every WORKER_HEARTBEAT seconds, workers announce their presence ' + 'to the master.',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_timeout_factor'),
            metavar='FACTOR',
            type=float,
            help='Scaling factor for heartbeat timeouts. '
            + "If the master doesn't hear from a worker in WORKER_HEARTBEAT*FACTOR, "
            + "the worker is assumed to have crashed. If a worker doesn't hear from "
            + "the master in MASTER_HEARTBEAT*FACTOR seconds, the master is assumed "
            + "to have crashed. Both cases result in shutdown. ",
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_startup_timeout'),
            metavar='STARTUP_TIMEOUT',
            type=float,
            help='Amount of time (in seconds) to wait for communication between '
            + 'the master and at least one worker. This may need to be changed '
            + 'on very large, heavily-loaded computer systems that start all processes '
            + 'simultaneously. ',
        )
        wm_group.add_argument(
            wmenv.arg_flag('zmq_shutdown_timeout'),
            metavar='SHUTDOWN_TIMEOUT',
            type=float,
            help='Amount of time (in seconds) to wait for workers to shut down.',
        )

    @classmethod
    def from_environ(cls, wmenv=None):
        if wmenv is None:
            wmenv = work_managers.environment.default_env

        # determine mode
        mode = wmenv.get_val('zmq_mode', 'master').lower()
        if mode in {'master', 'server'}:
            mode = 'master'
        elif mode in {'node', 'client'}:
            mode = 'node'
        else:
            raise ValueError('invalid ZMQ work manager mode {!r}'.format(mode))

        # determine number of workers
        # 0 with mode=='master' is a dedicated master
        # 0 with mode=='node' is a dedicated communications process (former ZMQRouter)
        n_workers = wmenv.get_val('n_workers', multiprocessing.cpu_count(), int)

        # We set this at the class level, because outside of testing, a node either
        # can support IPC or it can't, and there is no obvious need (currently)
        # to support both modes on an instance-by-instance basis
        comm_mode = wmenv.get_val('zmq_comm_mode', cls.default_comm_mode)
        ZMQWorkManager.internal_transport = comm_mode
        ZMQWorker.internal_transport = comm_mode
        ZMQNode.internal_transport = comm_mode

        write_host_info = wmenv.get_val('zmq_write_host_info')
        read_host_info = wmenv.get_val('zmq_read_host_info')
        master_heartbeat = wmenv.get_val('zmq_master_heartbeat', cls.default_master_heartbeat, float)
        worker_heartbeat = wmenv.get_val('zmq_worker_heartbeat', cls.default_worker_heartbeat, float)
        timeout_factor = wmenv.get_val('zmq_timeout_factor', cls.default_timeout_factor, float)
        startup_timeout = wmenv.get_val('zmq_startup_timeout', cls.default_startup_timeout, float)

        if mode == 'master':
            instance = ZMQWorkManager(n_workers)
        else:  # mode =='node'

            upstream_info = {}
            if read_host_info:
                upstream_info.update(cls.read_host_info(read_host_info))
            log.debug('upstream_info: {!r}'.format(upstream_info))

            upstream_rr_endpoint = wmenv.get_val('zmq_upstream_rr_endpoint', upstream_info.get('rr_endpoint'))
            upstream_ann_endpoint = wmenv.get_val('zmq_upstream_ann_endpoint', upstream_info.get('ann_endpoint'))

            if not (upstream_rr_endpoint and upstream_ann_endpoint):
                raise ZMQWMEnvironmentError('at least one upstream endpoint unspecified')

            # expand hostnames, if present, to IP addresses
            # reject wildcard hostnames, which is a logic error (can't connect to a host
            # without specifying an address)
            upstream_rr_endpoint = cls.canonicalize_endpoint(upstream_rr_endpoint, allow_wildcard_host=False)
            upstream_ann_endpoint = cls.canonicalize_endpoint(upstream_ann_endpoint, allow_wildcard_host=False)

            log.debug('upstream_rr_endpoint = {}'.format(upstream_rr_endpoint))
            log.debug('upstream_ann_endpoint = {}'.format(upstream_ann_endpoint))

            instance = ZMQNode(
                upstream_ann_endpoint=upstream_ann_endpoint, upstream_rr_endpoint=upstream_rr_endpoint, n_local_workers=n_workers
            )

        # Both server and node bind downstream endpoints, so that users get fan-out communications
        # "for free" when starting up a computational node
        downstream_rr_endpoint = cls.canonicalize_endpoint(
            wmenv.get_val('zmq_downstream_rr_endpoint', 'tcp://*:{}'.format(randport()))
        )
        downstream_ann_endpoint = cls.canonicalize_endpoint(
            wmenv.get_val('zmq_downstream_ann_endpoint', 'tcp://*:{}'.format(randport()))
        )
        instance.downstream_rr_endpoint = downstream_rr_endpoint
        instance.downstream_ann_endpoint = downstream_ann_endpoint

        instance.master_beacon_period = master_heartbeat
        instance.worker_beacon_period = worker_heartbeat
        instance.timeout_factor = timeout_factor
        instance.startup_timeout = startup_timeout

        assert isinstance(instance, IsNode)
        for worker in instance.local_workers:
            worker.master_beacon_period = master_heartbeat
            worker.worker_beacon_period = worker_heartbeat
            worker.timeout_factor = timeout_factor
            worker.startup_timeout = startup_timeout

        # We always write host info (since we are always either master or node)
        # we choose not to in the special case that read_host_info is '' but not None
        # (None implies nothing found on command line or in environment variables, but ''
        # implies that it was found somewhere but it is empty)
        if write_host_info is not None and write_host_info != '':
            instance.write_host_info(write_host_info)

        log.debug('prepared {!r} with:'.format(instance))
        log.debug('n_workers = {}'.format(n_workers))
        for attr in (
            'master_beacon_period',
            'worker_beacon_period',
            'startup_timeout',
            'timeout_factor',
            'downstream_rr_endpoint',
            'downstream_ann_endpoint',
        ):
            log.debug('{} = {!r}'.format(attr, getattr(instance, attr)))

        return instance

    @classmethod
    def read_host_info(cls, filename):
        return json.load(open(filename, 'rt'))

    @classmethod
    def canonicalize_endpoint(cls, endpoint, allow_wildcard_host=True):
        if endpoint.startswith('ipc://'):
            return endpoint
        elif endpoint.startswith('tcp://'):
            fields = endpoint[6:].split(':')

            # get IP address
            if fields[0] != '*':
                ipaddr = socket.gethostbyname(fields[0])
            else:
                if allow_wildcard_host:
                    ipaddr = '*'
                else:
                    raise ValueError('wildcard host not permitted')

            # get/generate port
            try:
                port = fields[1]
            except IndexError:
                # no port given; select one
                port = randport()
            else:
                port = int(fields[1])

            return 'tcp://{}:{}'.format(ipaddr, port)
        else:
            raise ValueError('unrecognized/unsupported endpoint: {!r}'.format(endpoint))

    def __init__(self, n_local_workers=1):
        ZMQCore.__init__(self)
        WorkManager.__init__(self)
        IsNode.__init__(self, n_local_workers)

        # Futures indexed by task ID
        self.futures = dict()

        # Tasks pending distribution
        self.outgoing_tasks = deque()

        # Tasks being processed by workers (indexed by worker_id)
        self.assigned_tasks = dict()

        # Identity information and last contact from workers
        self.worker_information = dict()  # indexed by worker_id
        self.worker_timeouts = PassiveMultiTimer()  # indexed by worker_id

        # Number of seconds between checks to see which workers have timed out
        self.worker_timeout_check = 5.0

        # Amount of time to wait for stray requests to arrive so that workers shut down properly
        self.shutdown_timeout = 0.5

        self.master_id = self.node_id

    @property
    def n_workers(self):
        return len(self.worker_information)

    def submit(self, fn, args=None, kwargs=None):
        if self.futures is None:
            # We are shutting down
            raise ZMQWMEnvironmentError('work manager is shutting down')
        future = WMFuture()
        task = Task(fn, args or (), kwargs or {}, task_id=future.task_id)
        self.futures[task.task_id] = future
        self.outgoing_tasks.append(task)
        # Wake up the communications loop (if necessary) to announce new tasks
        self.send_inproc_message(Message.TASKS_AVAILABLE)
        return future

    def submit_many(self, tasks):
        if self.futures is None:
            # We are shutting down
            raise ZMQWMEnvironmentError('work manager is shutting down')
        futures = []
        for (fn, args, kwargs) in tasks:
            future = WMFuture()
            task = Task(fn, args, kwargs, task_id=future.task_id)
            self.futures[task.task_id] = future
            self.outgoing_tasks.append(task)
            futures.append(future)
        # Wake up the communications loop (if necessary) to announce new tasks
        self.send_inproc_message(Message.TASKS_AVAILABLE)
        return futures

    def send_message(self, socket, message, payload=None, flags=0):
        message = Message(message, payload)
        message.master_id = self.node_id
        super().send_message(socket, message, payload, flags)

    def handle_result(self, socket, msg):
        self.send_ack(socket, msg)
        with self.message_validation(msg):
            assert msg.message == Message.RESULT
            assert isinstance(msg.payload, Result)
            assert msg.payload.task_id in self.futures
            assert self.assigned_tasks[msg.src_id].task_id == msg.payload.task_id

        result = msg.payload

        future = self.futures.pop(result.task_id)
        del self.assigned_tasks[msg.src_id]
        if result.exception is not None:
            future._set_exception(result.exception, result.traceback)
        else:
            future._set_result(result.result)

    def handle_task_request(self, socket, msg):
        if not self.outgoing_tasks:
            # No tasks available
            self.send_nak(socket, msg)
        else:
            task = self.outgoing_tasks.popleft()

            worker_id = msg.src_id
            self.assigned_tasks[worker_id] = task

            self.send_message(socket, Message.TASK, task)

    def update_worker_information(self, msg):
        if msg.message == Message.IDENTIFY:
            with self.message_validation(msg):
                assert isinstance(msg.payload, dict)
            self.worker_information[msg.src_id] = msg.payload
        else:
            self.worker_information[msg.src_id] = {}

        try:
            self.worker_timeouts.reset(msg.src_id)
        except KeyError:
            self.worker_timeouts.add_timer(msg.src_id, self.worker_beacon_period * self.timeout_factor)

    def check_workers(self):
        expired_worker_ids = self.worker_timeouts.which_expired()
        for expired_worker_id in expired_worker_ids:
            try:
                worker_description = '{!s} ({!s})'.format(
                    expired_worker_id, self.worker_information[expired_worker_id]['description']
                )
            except KeyError:
                worker_description = str(expired_worker_id)

            self.log.error('no contact from worker {}: {}'.format(expired_worker_id, worker_description))

            self.remove_worker(expired_worker_id)

    def remove_worker(self, worker_id):
        try:
            expired_task = self.assigned_tasks.pop(worker_id)
        except KeyError:
            pass
        else:
            self.log.error('aborting task {!r} running on expired worker {!s}'.format(expired_task, worker_id))
            future = self.futures.pop(expired_task.task_id)
            future._set_exception(ZMQWorkerMissing('worker running this task disappeared'))
        del self.worker_information[worker_id]

    def shutdown_clear_tasks(self):
        '''Abort pending tasks with error on shutdown.'''
        while self.futures:
            task_id, future = self.futures.popitem()
            future._set_exception(ZMQWMEnvironmentError('work manager shut down during task'))
        self.futures = None

    def comm_loop(self):
        self.context = zmq.Context()

        rr_socket = self.context.socket(zmq.REP)
        ann_socket = self.context.socket(zmq.PUB)

        for endpoint in (self.local_rr_endpoint, self.downstream_rr_endpoint):
            if endpoint:
                rr_socket.bind(endpoint)

        for endpoint in (self.local_ann_endpoint, self.downstream_ann_endpoint):
            if endpoint:
                ann_socket.bind(endpoint)

        inproc_socket = self.context.socket(zmq.SUB)
        inproc_socket.setsockopt(zmq.SUBSCRIBE, b'')
        inproc_socket.bind(self.inproc_endpoint)

        poller = zmq.Poller()
        poller.register(inproc_socket, zmq.POLLIN)
        poller.register(rr_socket, zmq.POLLIN)

        timers = PassiveMultiTimer()
        timers.add_timer('tasks_avail', self.master_beacon_period)
        timers.add_timer('master_beacon', self.master_beacon_period)
        timers.add_timer('worker_timeout_check', self.worker_beacon_period * self.timeout_factor)
        timers.add_timer('startup_timeout', self.startup_timeout)
        timers.reset()

        self.log.debug('master beacon period: {!r}'.format(self.master_beacon_period))
        self.log.debug('startup timeout: {!r}'.format(self.startup_timeout))

        peer_found = False

        try:
            # Send a master alive message immediately; it will get discarded if necessary
            self.send_message(ann_socket, Message.MASTER_BEACON)

            while True:
                # If a timer is already expired, next_expiration_in() will return 0, which
                # zeromq interprets as infinite wait; so instead we select a 1 ms wait in this
                # case.
                timeout = (timers.next_expiration_in() or 0.001) * 1000
                # Wake up every second to check for signals
                timeout = min(timeout, 1000)
                poll_results = dict(poller.poll(timeout))

                if inproc_socket in poll_results:
                    msgs = self.recv_all(inproc_socket, validate=False)
                    # Check for shutdown; do nothing else if shutdown is signalled
                    if Message.SHUTDOWN in (msg.message for msg in msgs):
                        self.log.debug('shutdown received')
                        break
                    # Check for any other wake-up messages
                    for msg in msgs:
                        if msg.message == Message.TASKS_AVAILABLE:
                            self.send_message(ann_socket, Message.TASKS_AVAILABLE)

                if rr_socket in poll_results:
                    msg = self.recv_message(rr_socket)
                    self.update_worker_information(msg)

                    if msg.message == Message.TASK_REQUEST:
                        self.handle_task_request(rr_socket, msg)
                    elif msg.message == Message.RESULT:
                        self.handle_result(rr_socket, msg)
                    else:
                        self.send_ack(rr_socket, msg)

                    if self.worker_information:
                        peer_found = True

                if timers.expired('tasks_avail'):
                    if self.outgoing_tasks:
                        self.send_message(ann_socket, Message.TASKS_AVAILABLE)
                    timers.reset('tasks_avail')

                if timers.expired('master_beacon'):
                    self.send_message(ann_socket, Message.MASTER_BEACON)
                    timers.reset('master_beacon')

                if peer_found and timers.expired('worker_timeout_check'):
                    self.check_workers()
                    if not self.worker_information:
                        self.log.error('all workers disappeared; exiting')
                        break
                    timers.reset('worker_timeout_check')

                if not peer_found and timers.expired('startup_timeout'):
                    self.log.error('startup phase elapsed with no contact from workers; shutting down')
                    while self.futures:
                        future = self.futures.popitem()[1]
                        future._set_exception(ZMQWorkerMissing('no workers available'))
                    break

            # Post a shutdown message
            self.log.debug('sending shutdown on ann_socket')
            self.send_message(ann_socket, Message.SHUTDOWN)
            poller.unregister(inproc_socket)

            # Clear tasks
            self.shutdown_clear_tasks()

            # Clear incoming queue of requests, to let clients exit request/reply states gracefully
            # (clients will still timeout in these states if necessary)
            timers.add_timer('shutdown', self.shutdown_timeout)
            while not timers.expired('shutdown'):
                poll_results = dict(poller.poll(self.shutdown_timeout / 10 * 1000))
                if rr_socket in poll_results:
                    msg = self.recv_message(rr_socket)
                    self.send_nak(rr_socket, msg)

        finally:
            self.context.destroy(linger=1)
            self.context = None
            self.remove_ipc_endpoints()

    def startup(self):
        IsNode.startup(self)
        super().startup()

    def shutdown(self):
        self.signal_shutdown()
        IsNode.shutdown(self)
        self.join()
        super().shutdown()



import h5py
import numpy as np

from westpa.tools import WESTTool, WESTDataReader, IterRangeSelection, ProgressIndicatorComponent

import westpa
from westpa.core import h5io
from westpa.core.data_manager import seg_id_dtype, n_iter_dtype, weight_dtype
from westpa.core.binning import assignments_list_to_table


class WNTopTool(WESTTool):
    prog = 'w_ntop'
    description = '''\
Select walkers from bins . An assignment file mapping walkers to
bins at each timepoint is required (see``w_assign --help`` for further
information on generating this file). By default, high-weight walkers are
selected (hence the name ``w_ntop``: select the N top-weighted walkers from
each bin); however, minimum weight walkers and randomly-selected walkers
may be selected instead.


-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "ntop.h5") contains the following
datasets:

  ``/n_iter`` [iteration]
    *(Integer)* Iteration numbers for each entry in other datasets.

  ``/n_segs`` [iteration][bin]
    *(Integer)* Number of segments in each bin/state in the given iteration.
    This will generally be the same as the number requested with
    ``--n/--count`` but may be smaller if the requested number of walkers
    does not exist.

  ``/seg_ids`` [iteration][bin][segment]
    *(Integer)* Matching segments in each iteration for each bin.
    For an iteration ``n_iter``, only the first ``n_iter`` entries are
    valid. For example, the full list of matching seg_ids in bin 0 in the
    first stored iteration is ``seg_ids[0][0][:n_segs[0]]``.

  ``/weights`` [iteration][bin][segment]
    *(Floating-point)* Weights for each matching segment in ``/seg_ids``.


-----------------------------------------------------------------------------
Command-line arguments
-----------------------------------------------------------------------------
'''

    def __init__(self):
        super().__init__()

        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection()
        self.progress = ProgressIndicatorComponent()
        self.output_file = None
        self.assignments_filename = None
        self.output_filename = None
        self.what = None
        self.timepoint = None
        self.count = None

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.iter_range.add_args(parser)

        igroup = parser.add_argument_group('input options')
        igroup.add_argument(
            '-a',
            '--assignments',
            default='assign.h5',
            help='''Use assignments from the given ASSIGNMENTS file (default: %(default)s).''',
        )

        sgroup = parser.add_argument_group('selection options')
        sgroup.add_argument(
            '-n',
            '--count',
            type=int,
            default=1,
            help='''Select COUNT walkers from each iteration for each bin (default: %(default)s).''',
        )
        sgroup.add_argument(
            '-t',
            '--timepoint',
            type=int,
            default=-1,
            help='''Base selection on the given TIMEPOINT within each iteration. Default (-1)
                            corresponds to the last timepoint.''',
        )
        cgroup = parser.add_mutually_exclusive_group()
        cgroup.add_argument(
            '--highweight',
            dest='select_what',
            action='store_const',
            const='highweight',
            help='''Select COUNT highest-weight walkers from each bin.''',
        )
        cgroup.add_argument(
            '--lowweight',
            dest='select_what',
            action='store_const',
            const='lowweight',
            help='''Select COUNT lowest-weight walkers from each bin.''',
        )
        cgroup.add_argument(
            '--random',
            dest='select_what',
            action='store_const',
            const='random',
            help='''Select COUNT walkers randomly from each bin.''',
        )
        parser.set_defaults(select_what='highweight')

        ogroup = parser.add_argument_group('output options')
        ogroup.add_argument('-o', '--output', default='ntop.h5', help='''Write output to OUTPUT (default: %(default)s).''')
        self.progress.add_args(parser)

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        with self.data_reader:
            self.iter_range.process_args(args)
        self.what = args.select_what
        self.output_filename = args.output
        self.assignments_filename = args.assignments
        self.count = args.count
        self.timepoint = args.timepoint

    def go(self):
        self.data_reader.open('r')
        assignments_file = h5py.File(self.assignments_filename, mode='r')
        output_file = h5io.WESTPAH5File(self.output_filename, mode='w')
        pi = self.progress.indicator
        count = self.count
        timepoint = self.timepoint

        nbins = assignments_file.attrs['nbins'] + 1
        assignments_ds = assignments_file['assignments']

        iter_start, iter_stop = self.iter_range.iter_start, self.iter_range.iter_stop
        iter_count = iter_stop - iter_start
        h5io.check_iter_range_least(assignments_ds, iter_start, iter_stop)
        nsegs = assignments_file['nsegs'][h5io.get_iteration_slice(assignments_file['nsegs'], iter_start, iter_stop)]

        output_file.create_dataset('n_iter', dtype=n_iter_dtype, data=list(range(iter_start, iter_stop)))

        seg_count_ds = output_file.create_dataset('nsegs', dtype=np.uint, shape=(iter_count, nbins))
        matching_segs_ds = output_file.create_dataset(
            'seg_ids',
            shape=(iter_count, nbins, count),
            dtype=seg_id_dtype,
            chunks=h5io.calc_chunksize((iter_count, nbins, count), seg_id_dtype),
            shuffle=True,
            compression=9,
        )
        weights_ds = output_file.create_dataset(
            'weights',
            shape=(iter_count, nbins, count),
            dtype=weight_dtype,
            chunks=h5io.calc_chunksize((iter_count, nbins, count), weight_dtype),
            shuffle=True,
            compression=9,
        )
        what = self.what

        with pi:
            pi.new_operation('Finding matching segments', extent=iter_count)
            for iiter, n_iter in enumerate(range(iter_start, iter_stop)):
                assignments = np.require(
                    assignments_ds[h5io.get_iteration_entry(assignments_ds, n_iter) + np.index_exp[:, timepoint]],
                    dtype=westpa.core.binning.index_dtype,
                )
                all_weights = self.data_reader.get_iter_group(n_iter)['seg_index']['weight']

                # the following Cython function just executes this loop:
                # for iseg in xrange(nsegs[iiter]):
                #    segs_by_bin[iseg,assignments[iseg]] = True
                segs_by_bin = assignments_list_to_table(nsegs[iiter], nbins, assignments)
                for ibin in range(nbins):
                    segs = np.nonzero(segs_by_bin[:, ibin])[0]

                    seg_count_ds[iiter, ibin] = min(len(segs), count)

                    if len(segs):
                        weights = all_weights.take(segs)

                        if what == 'lowweight':
                            indices = np.argsort(weights)[:count]
                        elif what == 'highweight':
                            indices = np.argsort(weights)[::-1][:count]
                        else:
                            assert what == 'random'
                            indices = np.random.permutation(len(weights))

                        matching_segs_ds[iiter, ibin, : len(segs)] = segs.take(indices)
                        weights_ds[iiter, ibin, : len(segs)] = weights.take(indices)
                        del segs, weights

                del assignments, segs_by_bin, all_weights
                pi.progress += 1


def entry_point():
    WNTopTool().main()


if __name__ == '__main__':
    entry_point()



"""
A work manager which uses MPI to distribute tasks and collect results.
"""

import sys
import time
import traceback
import logging
import threading
from collections import deque
from mpi4py import MPI
from westpa.work_managers import WorkManager, WMFuture

log = logging.getLogger(__name__)


# +------+
# | Task |
# +------+
class Task:
    """Tasks are tuples of (task_id, function, args, keyword args)"""

    def __init__(self, task_id, fn, args, kwargs):
        self.task_id = task_id
        self.fn = fn
        self.args = args if args is not None else ()
        self.kwargs = kwargs if kwargs is not None else {}

    def __repr__(self):
        return '<Task {self.task_id}: {self.fn!r}(*{self.args!r}, **{self.kwargs!r})>'.format(self=self)


# +----------------+
# | MPIWorkManager |
# +----------------+
class MPIWorkManager(WorkManager):
    """MPIWorkManager factory."""

    @classmethod
    def from_environ(cls, wmenv=None):
        return cls()

    def __new__(cls):
        """Creates a Serial WorkManager if size is 1.  Otherwise creates a
        single Manager and size-1 Worker.
        """
        log.debug('MPIWorkManager.__new__()')
        assert MPI.Is_initialized()
        assert MPI.Is_thread_main()

        rank = MPI.COMM_WORLD.Get_rank()
        size = MPI.COMM_WORLD.Get_size()

        if size == 1:
            return super().__new__(Serial)
        elif rank == 0:
            return super().__new__(Manager)
        else:
            return super().__new__(Worker)

    def __init__(self):
        """Initialize info shared by Manager and Worker classes."""
        log.debug('MPIWorkManager.__init__()')

        super().__init__()
        comm = MPI.COMM_WORLD
        self.comm = MPI.COMM_WORLD
        self.rank = comm.Get_rank()
        self.size = comm.Get_size()
        self.name = MPI.Get_processor_name()

        # some tags
        self.task_tag = 110  # tag for server to client msgs
        self.result_tag = 120  # tag for client to server msgs
        self.shutdown_tag = 130  # tag for server to client to stop

        self.managerID = 0

    def submit(self, fn, args=None, kwargs=None):
        """Adhere to WorkManager interface.  This method should never be
        called.
        """
        assert False


# +--------+
# | Serial |
# +--------+
# TODO: no need for the code replication here, just use the original serial wm
class Serial(MPIWorkManager):
    """Replication of the serial work manager.  This is a fallback for MPI runs
    that request only 1 (size=1) processor.
    """

    def __init__(self):
        super().__init__()
        log.debug('Serial.__init__()')

    def submit(self, fn, args=None, kwargs=None):
        log.debug('Serial.__init__()')

        ft = WMFuture()
        try:
            result = fn(*(args if args is not None else ()), **(kwargs if kwargs is not None else {}))
        except Exception as e:
            ft._set_exception(e, sys.exc_info()[2])
        else:
            ft._set_result(result)
        return ft


# +---------+
# | Manager |
# +---------+
class Manager(MPIWorkManager):
    """Manager of the MPIWorkManage.  Distributes tasks to Worker as they are
    received from the sim_manager.  In addition to the main thread, this class
    spawns two threads, a receiver and a dispatcher.
    """

    def __init__(self):
        """Initialize different state variables used by Manager."""
        super().__init__()
        log.debug('Manager__init__()')

        # number of workers
        self.nworkers = self.size - 1

        # list of worker ranks
        self.workerIDs = range(1, self.size)
        assert self.nworkers == len(self.workerIDs)

        # deque of idle workers
        self.dests = deque(self.workerIDs)

        # deque of tesks
        self.tasks = deque()

        # number of unmatched send/receive pairs
        self.nPending = 0

        # thread shutdown sentinel
        self.shutItDown = False

        # task_id, future key value pair
        self.pending_futures = dict()

        # list of manager threads
        self.workers = []

        # thread lock
        self.lock = threading.Lock()

    def startup(self):
        """Spawns the dispatcher and receiver threads."""
        log.debug('Manager.startup()')

        if not self.running:
            self.workers.append(threading.Thread(name='dispatcher', target=self._dispatcher))
            self.workers.append(threading.Thread(name='receiver', target=self._receiver))

            for t in self.workers:
                t.start()
                log.info('Started thread: %s' % t.getName())

            self.running = True

    def _dispatcher(self):
        """Continuously dispatches tasks to idle destinations until the
        shutdown sentinel is set.
        """
        log.debug('Manager._dispatcher()')
        assert MPI.Is_thread_main() is False
        assert threading.currentThread().getName() == "dispatcher"

        while not self.shutItDown:

            req = []
            # do we have work and somewhere to send it?
            while self.tasks and self.dests:

                with self.lock:
                    task = self.tasks.popleft()
                    sendTo = self.dests.popleft()
                    self.nPending += 1

                req.append(self.comm.isend(task, dest=sendTo, tag=self.task_tag))

            # make sure all sends completed
            MPI.Request.Waitall(requests=req)

            # force context switching ( 1ms )
            time.sleep(0.001)

    def _receiver(self):
        """Continuously receives futures from workers until the shutdown
        sentinel is set.
        """
        log.debug('Manager._receiver()')
        assert MPI.Is_thread_main() is False
        assert threading.currentThread().getName() == "receiver"

        while not self.shutItDown:

            # are we waiting on any results?
            while self.nPending:

                stat = MPI.Status()
                (tid, msg, val) = self.comm.recv(source=MPI.ANY_SOURCE, tag=self.result_tag, status=stat)
                log.debug('Manager._receiver received task: %s' % tid)

                # update future
                ft = self.pending_futures.pop(tid)
                if msg == 'exception':
                    ft._set_exception(*val)
                else:
                    ft._set_result(val)

                with self.lock:
                    self.dests.append(stat.Get_source())
                    self.nPending -= 1

            # force context switching ( 1ms )
            time.sleep(0.001)

    def submit(self, fn, args=None, kwargs=None):
        """Receive task from simulation manager and add it to pending_futures."""
        log.debug('Manager.submit()')

        ft = WMFuture()
        task_id = ft.task_id
        with self.lock:
            self.tasks.append(Task(task_id, fn, args, kwargs))
            self.pending_futures[task_id] = ft

        return ft

    def shutdown(self):
        """Send shutdown tag to all worker processes, and set the shutdown
        sentinel to stop the receiver and dispatcher loops.
        """
        log.debug('Manager.shutdown()')

        # wait on any unfinished work
        while self.pending_futures:
            pass

        # send shutdown msg to all workers
        req = [MPI.REQUEST_NULL] * self.nworkers
        for rank in self.workerIDs:
            req[rank - 1] = self.comm.isend(MPI.BOTTOM, dest=rank, tag=self.shutdown_tag)

        MPI.Request.Waitall(requests=req)

        # stop threads
        self.shutItDown = True

        for t in self.workers:
            t.join()
            log.info('Stopped thread: %s' % t.getName())

        self.running = False


# +--------+
# | Worker |
# +--------+
class Worker(MPIWorkManager):
    """Client class for executing tasks as distributed by the Manager in the
    MPI Work Manager
    """

    def __init__(self):
        super().__init__()
        log.debug('Worker.__init__() %s' % self.rank)

    def startup(self):
        """Clock the worker in for work."""
        log.debug('Worker.startup() %s' % self.rank)
        if not self.running:

            self.clockIn()

            self.running = True

    def clockIn(self):
        """Do each task as it comes in.  The completion of a task is
        notice to the manager that more work is welcome.
        """
        log.info('Worker %s clocking in.' % self.rank)

        comm = self.comm

        while True:

            stat = MPI.Status()
            task = comm.recv(source=self.managerID, tag=MPI.ANY_TAG, status=stat)

            tag = stat.Get_tag()

            if tag == self.task_tag:

                log.debug('Worker %s received task: %s' % (self.rank, task.task_id))

                # do the work
                try:
                    rv = task.fn(*task.args, **task.kwargs)
                except BaseException as e:
                    ro = (task.task_id, 'exception', (e, traceback.format_exc()))
                else:
                    ro = (task.task_id, 'result', rv)

                # send result back to manager
                comm.send(ro, dest=self.managerID, tag=self.result_tag)

            if tag == self.shutdown_tag:
                log.info('Worker %s clocking out.' % self.rank)
                return

    @property
    def is_master(self):
        """Worker processes need to be marked as not manager.  This ensures that
        the proper branching is followed in w_run.py.
        """
        return False



from westpa.tools import WESTMasterCommand, WESTParallelTool
from warnings import warn

from westpa.cli.tools.w_direct import DStateProbs

# Just a shim to make sure everything works and is backwards compatible.
# We're making sure it has the appropriate functions so that it can be called
# as a regular tool, and not a subcommand.


class WStateProbs(DStateProbs):
    subcommand = 'trace'
    help_text = 'averages and CIs for path-tracing kinetics analysis'
    default_output_file = 'stateprobs.h5'
    # This isn't strictly necessary, but for the moment, here it is.
    # We really need to modify the underlying class so that we don't pull this sort of stuff if it isn't necessary.
    # That'll take some case handling, which is fine.
    default_kinetics_file = 'assign.h5'


class WDirect(WESTMasterCommand, WESTParallelTool):
    prog = 'w_stateprobs'
    subcommands = [WStateProbs]
    subparsers_title = 'calculate state-to-state kinetics by tracing trajectories'
    description = '''\
Calculate average populations and associated errors in state populations from
weighted ensemble data. Bin assignments, including macrostate definitions,
are required. (See "w_assign --help" for more information).

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, usually "stateprobs.h5") contains the following
dataset:

  /avg_state_pops [state]
    (Structured -- see below) Population of each state across entire
    range specified.

If --evolution-mode is specified, then the following additional dataset is
available:

  /state_pop_evolution [window][state]
    (Structured -- see below). State populations based on windows of
    iterations of varying width.  If --evolution-mode=cumulative, then
    these windows all begin at the iteration specified with
    --start-iter and grow in length by --step-iter for each successive
    element. If --evolution-mode=blocked, then these windows are all of
    width --step-iter (excluding the last, which may be shorter), the first
    of which begins at iteration --start-iter.

The structure of these datasets is as follows:

  iter_start
    (Integer) Iteration at which the averaging window begins (inclusive).

  iter_stop
    (Integer) Iteration at which the averaging window ends (exclusive).

  expected
    (Floating-point) Expected (mean) value of the rate as evaluated within
    this window, in units of inverse tau.

  ci_lbound
    (Floating-point) Lower bound of the confidence interval on the rate
    within this window, in units of inverse tau.

  ci_ubound
    (Floating-point) Upper bound of the confidence interval on the rate
    within this window, in units of inverse tau.

  corr_len
    (Integer) Correlation length of the rate within this window, in units
    of tau.

Each of these datasets is also stamped with a number of attributes:

  mcbs_alpha
    (Floating-point) Alpha value of confidence intervals. (For example,
    *alpha=0.05* corresponds to a 95% confidence interval.)

  mcbs_nsets
    (Integer) Number of bootstrap data sets used in generating confidence
    intervals.

  mcbs_acalpha
    (Floating-point) Alpha value for determining correlation lengths.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''


def entry_point():
    warn('{} is being deprecated.  Please use w_direct instead.'.format(WDirect.prog))
    # If we're not really supporting subcommands...
    import sys

    try:
        if sys.argv[1] != 'trace':
            sys.argv.insert(1, 'trace')
    except Exception:
        sys.argv.insert(1, 'trace')
    WDirect().main()


if __name__ == '__main__':
    entry_point()



import argparse
import logging

import westpa


log = logging.getLogger('w_truncate')

warning_string = '''\
NOTE: w_truncate only deletes iteration groups from the HDF5 data store.
It is recommended that any iteration data saved to the file system (e.g. in the
traj_segs directory) is deleted or moved for the corresponding iterations.
'''


def entry_point():
    parser = argparse.ArgumentParser(
        'w_truncate',
        description='''\
    Remove all iterations after a certain point in a WESTPA simulation.
    ''',
        epilog=warning_string,
    )

    westpa.rc.add_args(parser)
    parser.add_argument('-n', '--iter', dest='n_iter', type=int, help='Truncate this iteration and those following.')
    args = parser.parse_args()
    westpa.rc.process_args(args, config_required=False)
    dm = westpa.rc.get_data_manager()
    dm.open_backing()
    # max_iter = dm.current_iteration
    n_iter = args.n_iter if args.n_iter > 0 else dm.current_iteration

    for i in range(n_iter, dm.current_iteration + 1):
        dm.del_iter_group(i)

    dm.del_iter_summary(n_iter)
    dm.current_iteration = n_iter - 1

    print('simulation data truncated after iteration {}'.format(dm.current_iteration))
    print('\n' + warning_string)

    dm.flush_backing()
    dm.close_backing()


if __name__ == '__main__':
    entry_point()



import logging

import numpy as np


import westpa
from westpa.oldtools.aframe import AnalysisMixin, ArgumentError

log = logging.getLogger(__name__)


class IterRangeMixin(AnalysisMixin):
    '''A mixin for limiting the range of data considered for a given analysis. This should go after
    DataManagerMixin'''

    def __init__(self):
        super().__init__()

        self.first_iter = None
        self.last_iter = None
        self.iter_step = 1

        include_args = self.include_args.setdefault('IterRangeMixin', {})
        include_args.setdefault('first_iter', True)
        include_args.setdefault('last_iter', True)
        include_args.setdefault('iter_step', True)

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        group = parser.add_argument_group('analysis range')
        if self.include_args['IterRangeMixin']['first_iter']:
            group.add_argument(
                '--start',
                '--begin',
                '--first',
                dest='first_iter',
                type=int,
                metavar='N_ITER',
                default=1,
                help='''Begin analysis at iteration N_ITER (default: %(default)d).''',
            )
        if self.include_args['IterRangeMixin']['last_iter']:
            group.add_argument(
                '--stop',
                '--end',
                '--last',
                dest='last_iter',
                type=int,
                metavar='N_ITER',
                help='''Conclude analysis with N_ITER, inclusive (default: last completed iteration).''',
            )
        if self.include_args['IterRangeMixin']['iter_step']:
            group.add_argument(
                '--step', dest='iter_step', type=int, metavar='STEP', help='''Analyze/report in blocks of STEP iterations.'''
            )

    def process_args(self, args, upcall=True):
        if self.include_args['IterRangeMixin']['first_iter']:
            self.first_iter = args.first_iter or 1
        if self.include_args['IterRangeMixin']['last_iter']:
            self.last_iter = args.last_iter
        if self.include_args['IterRangeMixin']['iter_step']:
            self.iter_step = args.iter_step or 1

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def check_iter_range(self):
        assert hasattr(self, 'data_manager') and self.data_manager is not None

        self.first_iter = int(max(self.first_iter, 1))
        if self.last_iter is None or self.last_iter > self.data_manager.current_iteration - 1:
            self.last_iter = int(self.data_manager.current_iteration - 1)

        if self.first_iter == self.last_iter:
            raise ArgumentError('first and last iterations are the same')

        westpa.rc.pstatus(
            'Processing iterations from {self.first_iter:d} to {self.last_iter:d}, inclusive (step size {self.iter_step:d})'.format(
                self=self
            )
        )

    def iter_block_iter(self):
        '''Return an iterable of (block_first,block_last+1) over the blocks of iterations
        selected by --first/--last/--step.  NOTE WELL that the second of the pair follows Python
        iterator conventions and returns one past the last element of the block.'''

        for blkfirst in range(self.first_iter, self.last_iter + 1, self.iter_step):
            yield (blkfirst, min(self.last_iter, blkfirst + self.iter_step - 1) + 1)

    def n_iter_blocks(self):
        '''Return the number of blocks of iterations (as returned by ``iter_block_iter``) selected by --first/--last/--step.'''
        npoints = self.last_iter - self.first_iter + 1
        if npoints % self.iter_step == 0:
            return npoints // self.iter_step
        else:
            return npoints // self.iter_step + 1

    def record_data_iter_range(self, h5object, first_iter=None, last_iter=None):
        '''Store attributes ``first_iter`` and ``last_iter`` on the given HDF5 object (group/dataset)'''
        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter
        h5object.attrs['first_iter'] = first_iter
        h5object.attrs['last_iter'] = last_iter

    def record_data_iter_step(self, h5object, iter_step=None):
        '''Store attribute ``iter_step`` on the given HDF5 object (group/dataset).'''
        iter_step = iter_step or self.iter_step
        h5object.attrs['iter_step'] = iter_step

    def check_data_iter_range_least(self, h5object, first_iter=None, last_iter=None):
        '''Check that the given HDF5 object contains (as denoted by its ``first_iter``/``last_iter`` attributes) at least the
        data range specified.'''
        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter

        obj_first_iter = h5object.attrs.get('first_iter')
        obj_last_iter = h5object.attrs.get('last_iter')

        return obj_first_iter <= first_iter and obj_last_iter >= last_iter

    def check_data_iter_range_equal(self, h5object, first_iter=None, last_iter=None):
        '''Check that the given HDF5 object contains per-iteration data for exactly the specified iterations (as denoted by the
        object's ``first_iter`` and ``last_iter`` attributes'''

        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter

        obj_first_iter = h5object.attrs.get('first_iter')
        obj_last_iter = h5object.attrs.get('last_iter')

        return obj_first_iter == first_iter and obj_last_iter == last_iter

    def check_data_iter_step_conformant(self, h5object, iter_step=None):
        '''Check that the given HDF5 object contains per-iteration data at an iteration stride suitable for extracting data
        with the given stride.  (In other words, is the given ``iter_step`` a multiple of the stride with
        which data was recorded.)'''

        iter_step = iter_step or self.iter_step
        obj_iter_step = h5object.attrs.get('iter_step')
        return obj_iter_step % iter_step == 0

    def check_data_iter_step_equal(self, h5object, iter_step=None):
        '''Check that the given HDF5 object contains per-iteration data at an iteration stride the same as
        that specified.'''
        iter_step = iter_step or self.iter_step
        obj_iter_step = h5object.attrs.get('iter_step')
        return obj_iter_step == iter_step

    def slice_per_iter_data(self, dataset, first_iter=None, last_iter=None, iter_step=None, axis=0):
        '''Return the subset of the given dataset corresponding to the given iteration range and stride. Unless
        otherwise specified, the first dimension of the dataset is the one sliced.'''

        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter
        iter_step = iter_step or self.iter_step

        ds_first_iter = dataset.attrs['first_iter']
        ds_last_iter = dataset.attrs['last_iter']
        ds_iter_step = dataset.attrs.get('iter_step', 1)

        if first_iter < ds_first_iter or last_iter > ds_last_iter or ds_iter_step % iter_step > 0:
            raise IndexError(
                'Cannot slice requested iterations [{:d},{:d}] (stride={:d}) from dataset {!r} with range [{:d},{:d}] (stride={:d}).'.format(
                    first_iter, last_iter, iter_step, dataset, ds_first_iter, ds_last_iter, ds_iter_step
                )
            )

        dimslices = []
        for idim in range(len(dataset.shape)):
            if idim == axis:
                dimslices.append(slice(first_iter - ds_first_iter, last_iter - ds_first_iter + iter_step, iter_step))
            else:
                dimslices.append(slice(None, None, None))

        dimslices = tuple(dimslices)
        log.debug('slicing {!r} with {!r}'.format(dataset, dimslices))
        data = dataset[dimslices]
        log.debug('resulting data is of shape {!r}'.format(data.shape))
        return data

    def iter_range(self, first_iter=None, last_iter=None, iter_step=None):
        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter
        iter_step = iter_step or self.iter_step

        return np.arange(first_iter, last_iter + 1, iter_step)



import warnings

import cython
import numpy as np
cimport numpy as np

ctypedef np.uint16_t index_t
ctypedef np.float64_t weight_t
ctypedef np.uint8_t bool_t
ctypedef np.int64_t seg_id_t
ctypedef np.uint_t uint_t # 32 bits on 32-bit systems, 64 bits on 64-bit systems

cdef double NAN = np.nan

weight_dtype = np.float64
index_dtype = np.uint16
bool_dtype = np.bool_

from westpa.core.binning.assign import UNKNOWN_INDEX as _UNKNOWN_INDEX
cdef index_t UNKNOWN_INDEX = _UNKNOWN_INDEX

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef flux_assign(np.ndarray[weight_t, ndim=1] weights,
                  np.ndarray[index_t, ndim=1] init_assignments,
                  np.ndarray[index_t, ndim=1] final_assignments,
                  np.ndarray[weight_t, ndim=2] flux_matrix):
    cdef:
        Py_ssize_t m,n
        index_t i, j
    n = len(weights)
    for m from 0 <= m < n:
        i = init_assignments[m]
        j = final_assignments[m]
        flux_matrix[i,j] += weights[m]
    return

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef pop_assign(np.ndarray[weight_t, ndim=1] weights,
                 np.ndarray[index_t, ndim=1] assignments,
                 np.ndarray[weight_t, ndim=1] populations):
    cdef:
        Py_ssize_t m,n
        index_t i,
    n = len(weights)
    for m from 0 <= m < n:
        i = assignments[m]
        populations[i] += weights[m]
    return

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef calc_rates(weight_t[:,::1] fluxes,
                 weight_t[::1] populations,
                 weight_t[:,::1] rates,
                 bool_t[:,::1] mask):
    '''Calculate a rate matrices from flux and population matrices. A matrix of the same
    shape as fluxes, is also produced, to be used for generating a mask for the rate
    matrices where initial state populations are zero.'''

    cdef:
        Py_ssize_t narrays, nbins
        index_t iarray, i, j

    nbins = fluxes.shape[0]

    with nogil:
        for i in range(nbins):
            if populations[i] == 0.0:
                for j in range(nbins):
                    mask[i,j] = 1
                    rates[i,j] = 0.0
            else:
                for j in range(nbins):
                    mask[i,j] = 0
                    rates[i,j] = fluxes[i,j] / populations[i]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef weight_t calculate_labeled_fluxes_alllags(Py_ssize_t nstates,
                                                weights,
                                                parent_ids,
                                                micro_assignments,
                                                traj_assignments,
                                                weight_t[:,:,:,:] fluxes) except 0.0:
    cdef:
        Py_ssize_t niters = len(weights), nsegs, npts
        weight_t twindow = 0.0
        long lastiter, firstiter, iiter, windowlen
        long seg_id, current_id, parent_id
        index_t ibin, ilabel, fbin, flabel
        weight_t weight
        weight_t[:] lweights

        index_t[:,:] lmicro
        index_t[:,:] ltraj

    # We need to trace backward in each window, so we go from end to beginning

    for lastiter in range(niters-1,-1,-1):
        for windowlen in range(1,niters+1):
            firstiter = lastiter-windowlen+1
            if firstiter < 0: continue

            # we loop over all trajectories that are alive as of the last iteration
            # in the averaging window
            lweights = weights[lastiter]
            lmicro = micro_assignments[lastiter]
            ltraj  = traj_assignments[lastiter]
            nsegs = lmicro.shape[0]
            npts =  lmicro.shape[1]

            for seg_id in range(nsegs):
                weight = lweights[seg_id]
                fbin = lmicro[seg_id,npts-1]
                flabel = ltraj[seg_id,npts-1]

                # trace upwards in history to firstiter
                iiter = lastiter
                current_id = seg_id
                parent_id = parent_ids[iiter][seg_id]
                while iiter > firstiter and parent_id >= 0:
                    iiter -= 1
                    current_id = parent_id
                    parent_id = parent_ids[iiter][current_id]

                ibin = micro_assignments[iiter][current_id][0]
                ilabel = traj_assignments[iiter][current_id][0]

                if ilabel >= nstates or flabel >= nstates:
                    raise ValueError('invalid state index (ilabel={},flabel={})'.format(ilabel,flabel))

                fluxes[ilabel,flabel,ibin,fbin] += weight
                twindow += weight*windowlen
    return twindow

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef weight_t calculate_labeled_fluxes(Py_ssize_t nstates,
                                        weights,
                                        parent_ids,
                                        micro_assignments,
                                        traj_assignments,
                                        weight_t[:,:,:,:] fluxes) except 0.0:
    cdef:
        Py_ssize_t niters = len(weights), nsegs, npts
        weight_t twindow = 0.0
        long lastiter, firstiter, iiter, windowlen
        long seg_id, current_id, parent_id
        index_t ibin, ilabel, fbin, flabel
        weight_t weight
        weight_t[:] lweights

        index_t[:,:] lmicro
        index_t[:,:] ltraj

    # we loop over all trajectories that are alive as of the last iteration
    # in the averaging window

    lastiter = niters-1
    windowlen = niters

    lweights = weights[lastiter]
    lmicro = micro_assignments[lastiter]
    ltraj  = traj_assignments[lastiter]
    nsegs = lmicro.shape[0]
    npts =  lmicro.shape[1]

    for seg_id in range(nsegs):
        weight = lweights[seg_id]
        fbin = lmicro[seg_id,npts-1]
        flabel = ltraj[seg_id,npts-1]

        # trace upwards in history to firstiter
        iiter = lastiter
        current_id = seg_id
        parent_id = parent_ids[iiter][seg_id]
        while iiter > 0 and parent_id >= 0:
            iiter -= 1
            current_id = parent_id
            parent_id = parent_ids[iiter][current_id]

        assert iiter == 0 or parent_id < 0
        assert 0 <= iiter < niters
        assert current_id >= 0

        ibin = micro_assignments[iiter][current_id][0]
        ilabel = traj_assignments[iiter][current_id][0]

        #if ilabel >= nstates or flabel >= nstates:
        #    raise ValueError('invalid state index (ilabel={},flabel={})'.format(ilabel,flabel))
        if ilabel < nstates and flabel < nstates:
            fluxes[ilabel,flabel,ibin,fbin] += weight
        twindow += weight*windowlen
    return twindow


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef object nested_to_flat_matrix(weight_t[:,:,:,:] input):
    '''Convert nested flux/rate matrix into a flat supermatrix.'''

    cdef:
        Py_ssize_t nstates = input.shape[0], nbins=input.shape[3], istate, ibin, jstate, jbin
        weight_t[:,:] _output

    output = np.empty((nstates*nbins,nstates*nbins), weight_dtype)
    _output = output

    for istate in range(nstates):
        for jstate in range(nstates):
            for ibin in range(nbins):
                for jbin in range(nbins):
                    #_output[istate*nbins+ibin, jstate*nbins+jbin] = input[istate, jstate, ibin, jbin]
                    _output[ibin*nstates+istate,jbin*nstates+jstate] = input[istate,jstate,ibin,jbin]

    return output

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef object nested_to_flat_vector(weight_t[:,:] input):
    '''Convert nested labeled population vector into a flat vector.'''

    cdef:
        Py_ssize_t nstates = input.shape[0], nbins=input.shape[1], istate, ibin
        weight_t[:] _output

    output = np.empty((nstates*nbins,), weight_dtype)
    _output = output

    for istate in range(nstates):
        for ibin in range(nbins):
            #_output[istate*nbins+ibin] = input[istate, ibin]
            _output[ibin*nstates+istate] = input[istate,ibin]

    return output

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef object flat_to_nested_matrix(Py_ssize_t nstates, Py_ssize_t nbins, weight_t[:,:] input):
    '''Convert flat supermatrix into nested matrix.'''

    cdef:
        Py_ssize_t istate, jstate, ibin, jbin
        weight_t[:,:,:,:] _output

    if input.shape[0] != nstates*nbins or input.shape[1] != nstates*nbins:
        # since input.shape is a C vector rather than a tuple, we can't print
        # it easily
        raise TypeError('input has incorrect shape for {} states and {} bins'.format(nstates, nbins))

    output = np.empty((nstates, nstates, nbins, nbins), weight_dtype)
    _output = output

    for istate in range(nstates):
        for jstate in range(nstates):
            for ibin in range(nbins):
                for jbin in range(nbins):
                    #_output[istate,jstate,ibin,jbin] = input[istate*nbins+ibin, jstate*nbins+jbin]
                    _output[istate,jstate,ibin,jbin] = input[ibin*nstates+istate,jbin*nstates+jstate]
    return output

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef object flat_to_nested_vector(Py_ssize_t nstates, Py_ssize_t nbins, weight_t[:] input):
    '''Convert flat "supervector" into nested vector.'''

    cdef:
        Py_ssize_t istate, ibin
        weight_t[:,:] _output

    if input.shape[0] != nstates*nbins:
        raise TypeError('input has incorrect shape for {} states and {} bins'.format(nstates, nbins))

    output = np.empty((nstates, nbins), weight_dtype)
    _output = output

    for istate in xrange(nstates):
        for ibin in xrange(nbins):
            #_output[istate,ibin] = input[istate*nbins+ibin]
            _output[istate,ibin] = input[ibin*nstates+istate]

    return output

@cython.boundscheck(True)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef _reduce_labeled_rate_matrix_to_macro(Py_ssize_t nstates, Py_ssize_t nbins, weight_t[:,:] rates, weight_t[:] pops):
    '''Reduce a labeled microstate rate matrix into a macrostate rate matrix. This is
    for internal use, where the rates/pops vectors have been blocked by state.'''

    cdef:
        Py_ssize_t istate, jstate, ibin, jbin
        weight_t[:,:] _macro_rates
        weight_t sspop, rate_elem, traj_ens_pop

    macro_rates = np.zeros((nstates, nstates), np.float64)
    _macro_rates = macro_rates

    for istate in xrange(nstates):
        for jstate in xrange(nstates):
            for ibin in xrange(nbins):
                for jbin in xrange(nbins):
                    sspop = pops[ibin*nstates+istate]
                    rateelem = rates[ibin*nstates+istate,jbin*nstates+jstate]
                    _macro_rates[istate,jstate] += sspop*rateelem

    # Normalize by total population in each trajectory ensemble
    for istate in xrange(nstates):
        #traj_ens_pop = pops[istate].sum()
        traj_ens_pop = 0
        for ibin in xrange(nbins):
            traj_ens_pop += pops[ibin*nstates+istate]

        for jstate in xrange(nstates):
            _macro_rates[istate, jstate] /=  traj_ens_pop

    return macro_rates


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef labeled_flux_to_rate(weight_t[:,:,:,:] labeled_fluxes, weight_t[:,:] labeled_pops, object output=None):
    '''Convert a labeled flux matrix and corresponding labeled bin populations to
    a labeled rate matrix.'''

    cdef:
        Py_ssize_t istate, jstate, ibin, jbin, nstates, nbins
        weight_t[:,:,:,:] _rates

    nstates = labeled_fluxes.shape[0]
    nbins = labeled_fluxes.shape[2]

    if output is None:
        output = np.empty_like(labeled_fluxes)
    _rates = output

    with nogil:
        for istate in xrange(nstates):
            for jstate in xrange(nstates):
                for ibin in xrange(nbins):
                    for jbin in xrange(nbins):
                        if labeled_pops[istate,ibin] == 0.0:
                            if labeled_fluxes[istate,jstate,ibin,jbin] > 0.0:
                                with gil:
                                    #raise ValueError('flux matrix entry nonzero but population zero')
                                    warnings.warn('flux matrix entry nonzero but population zero')

                            _rates[istate,jstate,ibin,jbin] = 0.0
                        else:
                            _rates[istate,jstate,ibin,jbin] = labeled_fluxes[istate,jstate,ibin,jbin] / labeled_pops[istate,ibin]
    return output

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef sequence_macro_flux_to_rate(weight_t[:] dataset, weight_t[:,:] pops, Py_ssize_t istate, Py_ssize_t jstate, bint pairwise=True, stride=None):
    '''Convert a sequence of macrostate fluxes and corresponding list of trajectory ensemble populations
    to a sequence of rate matrices.

    If the optional ``pairwise`` is true (the default), then rates are normalized according to the
    relative probability of the initial state among the pair of states (initial, final); this is
    probably what you want, as these rates will then depend only on the definitions of the states
    involved (and never the remaining states). Otherwise (``pairwise'' is false), the rates are
    normalized according the probability of the initial state among *all* other states.'''

    cdef:
        Py_ssize_t iiter, nstates, itersum
        weight_t[:] _rates, _fluxsum, _pairsum, _psum

    rates = np.zeros((dataset.shape[0]), dtype=weight_dtype)
    #rates = :W
    fluxsum = np.zeros((dataset.shape[0]), dtype=weight_dtype)
    psum = np.zeros((dataset.shape[0]), dtype=weight_dtype)
    pairsum = np.zeros((dataset.shape[0]), dtype=weight_dtype)
    _fluxsum = fluxsum
    _pairsum = pairsum
    _psum = psum
    _rates = rates

    # We want to modify this to be the SUM of fluxes up till this point, divided by the SUM of the population till then.
    with nogil:
        for iiter in xrange(dataset.shape[0]):
            if iiter == 0:
                # We need to catch if we haven't entered the istate or jstate yet.
                # Otherwise, we introduce a NaN into the calculation that we cannot recover from.
                if pairwise and (pops[0,istate] + pops[0,jstate]) != 0.0:
                    _psum[0] = pops[0, istate] / (pops[0,istate] + pops[0,jstate])
                else:
                    _psum[0] = pops[0, istate]
                _fluxsum[0] = dataset[0]
            else:
                if pairwise and (pops[iiter,istate] + pops[iiter,jstate]) != 0.0:
                    _psum[iiter] = (pops[iiter, istate] / (pops[iiter,istate] + pops[iiter,jstate])) + _psum[iiter-1]
                else:
                    _psum[iiter] = pops[iiter,istate] + _psum[iiter-1]
                _fluxsum[iiter] = dataset[iiter] + _fluxsum[iiter-1]
            if _psum[iiter] > 0 and _fluxsum[iiter] > 0:
                _rates[iiter] = _fluxsum[iiter] / _psum[iiter]
            else:
                _rates[iiter] = 0.0

    return rates[iiter]

"""
In the following ``state`` is a 5-tuple of the following arrays of doubles:
last_time[nsegs]
last_entries[nsegs,nstates]
last_exits[nsegs,nstates]
last_exits_td[nsegs,nstates]
last_completions[nsegs,nstates]


It is intended to be opaque the the calling routines
"""

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef _fast_transition_state_copy(Py_ssize_t iiter,
                                  Py_ssize_t nstates,
                                  seg_id_t[:] parent_ids,
                                  object last_state):
    cdef:
        bint has_last_state = 0
        Py_ssize_t nsegs, seg_id, parent_id
        double[:] _last_time, _prev_last_time
        double[:,:] _last_entries, _last_exits, _prev_last_entries, _prev_last_exits, _last_exits_td, _prev_last_exits_td
        double[:,:,:] _last_completions, _prev_last_completions


    nsegs = parent_ids.shape[0]

    last_time = np.empty((nsegs,), np.double)
    # Use nstates + 1 to account for possible unknown states
    last_entries = np.empty((nsegs,nstates+1), np.double)
    last_exits = np.empty((nsegs,nstates+1), np.double)
    last_exits_td = np.empty((nsegs,nstates+1), np.double)
    last_completions = np.empty((nsegs,nstates+1,nstates+1), np.double)

    _last_time = last_time
    _last_entries = last_entries
    _last_exits = last_exits
    _last_exits_td = last_exits_td
    _last_completions = last_completions

    has_last_state = (last_state is not None)

    if has_last_state:
        _prev_last_time = last_state[0]
        _prev_last_entries = last_state[1]
        _prev_last_exits = last_state[2]
        _prev_last_exits_td = last_state[3]
        _prev_last_completions = last_state[4]

    for seg_id in xrange(nsegs):
        parent_id = parent_ids[seg_id]

        if not has_last_state or parent_id < 0:
            _last_time[seg_id] = 0.0
            _last_entries[seg_id,:] = 0.0
            _last_exits[seg_id,:] = 0.0
            _last_exits_td[seg_id,:] = 0.0
            _last_completions[seg_id,:,:] = 0.0
        else:
            _last_time[seg_id] = _prev_last_time[parent_id]
            _last_entries[seg_id,:] = _prev_last_entries[parent_id,:]
            _last_exits[seg_id,:] = _prev_last_exits[parent_id,:]
            _last_exits_td[seg_id,:] = _prev_last_exits_td[parent_id,:]
            _last_completions[seg_id,:,:] = _prev_last_completions[parent_id,:,:]

    return (last_time, last_entries, last_exits, last_exits_td, last_completions)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef find_macrostate_transitions(Py_ssize_t nstates,
                                  weight_t[:] weights,
                                  index_t[:,:] label_assignments,
                                  index_t[:,:] state_assignments,
                                  double dt,
                                  object state,
                                  weight_t[:,:] macro_fluxes,
                                  uint_t[:,:] macro_counts,
                                  weight_t[:] target_fluxes,
                                  uint_t[:] target_counts,
                                  object durations):
    cdef:
        Py_ssize_t nsegs, npts, seg_id, ipt
        double itime, tm, t_ed
        double[:] _last_time
        double[:,:] _last_entries, _last_exits, _last_exits_td
        double[:,:,:] _last_completions
        index_t flabel, ilabel, iistate, slabel
        weight_t _weight
    """
    A cythoned function designed to track how long macrostate transitions take.  Requires the simulation
    to have already been binned and placed into macrostates, as appropriate.  Called by functions such as
    w_kinetics to generate kinetics information in the 'per-tau' format.

    Parameters
    ----------

    weights : weight_t
        Weights, typically from the main .h5 file of the simulation.  Likely called from the data reader
        of the calling function.
    label_assignments : index_t
        Macrostate label assignments, as compatible with those outputted by w_assign.  Bins are marked as
        states (or ignored) in a previous step.  Should be in the form  of a 'tag', or 'color'; in this
        dataset, once a walker has been marked with a macrostate, it does not lose the macrostate
        assigment, even upon leaving the appropriately defined state bin, until it enters another state bin.
    state_assignments : index_t
        Macrostate label assignments, but without any 'color' tagging.
    dt : double
        The number of timesteps of the system.
    state : object
       The output of _fast_transition_state_copy.
    macro_fluxes : weight_t
        An array that contains state to state fluxes.
    macro_counts : uint_t
        An array that contains the observed number of state to state fluxes.
    target_fluxes : weight_t
        An array that contains the fluxes into the target state from any state.
    target_counts : uint_t
        An array that contains the observed number of fluxes into the target state from any state.
    durations : list_like (object)
        A list containing the calculated duration information, including the iistate, flabel (state to state),
        event duration, the weight of all walkers involved, and all seg_ids.

    """


    nsegs = label_assignments.shape[0]
    npts = label_assignments.shape[1]

    _last_time = state[0]
    _last_entries = state[1]
    _last_exits = state[2]
    _last_exits_td = state[3]
    _last_completions = state[4]

    for seg_id in xrange(nsegs):
        itime = _last_time[seg_id]
        _weight = weights[seg_id]

        # transitions never occur between the (overlapping) end point of previous iteration and beginning of
        # current iteration, so it suffices to start looking at timepoint 1 (and backwards to timepoint 0)
        for ipt in range(1,npts):
            tm = itime + ipt*dt
            flabel = label_assignments[seg_id,ipt]
            ilabel = label_assignments[seg_id,ipt-1]
            slabel = state_assignments[seg_id,ipt]

            # if we have left our state transition barrier...
            if flabel == slabel:
                _last_exits_td[seg_id,flabel] = tm

            if flabel != ilabel:
                target_fluxes[flabel] += _weight
                target_counts[flabel] += 1
                _last_exits[seg_id,ilabel] = tm
                _last_entries[seg_id,flabel] = tm

                for iistate in xrange(nstates):
                    # if we have more recently returned to iistate than arrived at flabel from iistate,
                    # we note a new completed transition from iistate to flabel
                    # equality applies only for 0, which means we're counting an arrival from the
                    # state where the trajectory started
                    if _last_exits[seg_id, iistate] > 0 and _last_entries[seg_id,iistate] >= _last_completions[seg_id,iistate,flabel]:
                        macro_fluxes[iistate,flabel] += _weight
                        macro_counts[iistate,flabel] += 1
                        _last_completions[seg_id,iistate,flabel] = tm

                        # omit circular transitions (for now) because it causes the transition
                        # list to explode
                        if iistate != flabel:
                            t_ed = tm - _last_exits_td[seg_id,iistate]
                            durations.append((iistate,flabel,t_ed,_weight, seg_id))
        _last_time[seg_id] = tm


cdef class StreamingStats2D:
    '''Calculate mean and variance of a series of two-dimensional arrays of shape (nbins, nbins)
    using an online algorithm. The statistics are accumulated along what would be axis=0 if the
    input arrays were stacked vertically.

    This code has been adapted from:
    http://www.johndcook.com/skewness_kurtosis.html'''

    cdef weight_t[:,::1] _M1
    cdef weight_t[:,::1] _M2
    cdef uint_t[:,::1] _n
    cdef Py_ssize_t _sz0, _sz1

    def __init__(self, tuple shape):

        assert len(shape) == 2

        self._n = np.zeros(shape, dtype=np.uint)
        self._M1 = np.zeros(shape, dtype=weight_dtype)
        self._M2 = np.zeros(shape, dtype=weight_dtype)
        self._sz0, self._sz1 = shape

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def update(self, weight_t[:,::1] x, bool_t[:,::1] mask):
        '''Update the running set of statistics given

        Parameters
        ----------
        x : 2d ndarray
            values from a single observation
        mask : 2d ndarray
            A uint8 array to exclude entries from the accumulated statistics.
        '''

        cdef:
            index_t i, j
            int n1
            double delta, delta_n, term1

        assert x.shape[0] == mask.shape[0] == self._sz0
        assert x.shape[1] == mask.shape[1] == self._sz1

        with nogil:
            for i in range(self._sz0):
                for j in range(self._sz1):
                    if not mask[i,j]:
                        n1 = self._n[i,j]
                        self._n[i,j] += 1
                        delta = x[i,j] - self._M1[i,j]
                        delta_n = delta / self._n[i,j]
                        term1 = delta * delta_n * n1
                        self._M1[i,j] += delta_n
                        self._M2[i,j] += term1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __add__(StreamingStats2D self, StreamingStats2D other):
        cdef:
            index_t i, j
            int n1
            double delta, delta2
            StreamingStats2D combined

        combined = StreamingStats2D((self._sz0, self._sz1))

        for i in range(self._sz0):
            for j in range(self._sz1):
                combined._n[i,j] = self._n[i,j] + other._n[i,j]
                delta = other._M1[i,j] - self._M1[i,j]
                delta2 = delta * delta
                combined._M1[i,j] = (other._n[i,j]*other._M1[i,j] + self._n[i,j]*self._M1[i,j]) / combined._n[i,j]
                combined._M2[i,j] = other._M2[i,j] + self._M2[i,j] + (delta2 * self._n[i,j] * other._n[i,j]) / combined._n[i,j]

        return combined


    def __iadd__(StreamingStats2D self, StreamingStats2D other):
        combined = self + other
        self = combined
        return self


    property mean:
        def __get__(self):
            tmp = np.asarray(self._M1)
            return np.nan_to_num(tmp)

    property var:
        def __get__(self):
            tmp_m = np.asarray(self._M2)
            tmp_n = np.asarray(self._n)
            return np.nan_to_num(tmp_m / tmp_n)

    property n:
        def __get__(self):
            return np.asarray(self._n)

        def __set__(self, val):
            self._n = val[:]

    property M1:
        def __get__(self):
            return np.asarray(self._M1)

        def __set__(self, val):
            self._M1 = val[:]

    property M2:
        def __get__(self):
            return np.asarray(self._M2)

        def __set__(self, val):
            self._M2 = val[:]


cdef class StreamingStats1D:
    '''Calculate mean and variance of a series of one-dimensional arrays of shape (nbins,)
    using an online algorithm. The statistics are accumulated along what would be axis=0 if the
    input arrays were stacked vertically.

    This code has been adapted from:
    http://www.johndcook.com/skewness_kurtosis.html'''

    cdef weight_t[::1] _M1
    cdef weight_t[::1] _M2
    cdef uint_t[::1] _n
    cdef Py_ssize_t _sz0

    def __init__(self, int nbins):

        self._n = np.zeros((nbins,), dtype=np.uint)
        self._M1 = np.zeros((nbins,), dtype=weight_dtype)
        self._M2 = np.zeros((nbins,), dtype=weight_dtype)
        self._sz0 = nbins

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def update(self, weight_t[::1] x, bool_t[::1] mask):
        '''Update the running set of statistics given

        Parameters
        ----------
        x : 1d ndarray
            values from a single observation
        mask : 1d ndarray
            A uint8 array to exclude entries from the accumulated statistics.
        '''

        cdef:
            index_t i
            int n1
            double delta, delta_n, term1

        assert x.shape[0] == mask.shape[0] == self._sz0

        with nogil:
            for i in range(self._sz0):
                if not mask[i]:
                    n1 = self._n[i]
                    self._n[i] += 1
                    delta = x[i] - self._M1[i]
                    delta_n = delta / self._n[i]
                    term1 = delta * delta_n * n1
                    self._M1[i] += delta_n
                    self._M2[i] += term1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def __add__(StreamingStats1D self, StreamingStats1D other):
        cdef:
            index_t i
            int n1
            double delta, delta2
            StreamingStats1D combined

        combined = StreamingStats1D(self._sz0)

        for i in range(self._sz0):
            combined._n[i] = self._n[i] + other._n[i]
            delta = other._M1[i] - self._M1[i]
            delta2 = delta * delta
            combined._M1[i] = (other._n[i]*other._M1[i] + self._n[i]*self._M1[i]) / combined._n[i]
            combined._M2[i] = other._M2[i] + self._M2[i] + (delta2 * self._n[i] * other._n[i]) / combined._n[i]

        return combined


    def __iadd__(StreamingStats1D self, StreamingStats1D other):
        combined = self + other
        self = combined
        return self


    property mean:
        def __get__(self):
            tmp = np.asarray(self._M1)
            return np.nan_to_num(tmp)

    property var:
        def __get__(self):
            tmp_m = np.asarray(self._M2)
            tmp_n = np.asarray(self._n)
            return np.nan_to_num(tmp_m / tmp_n)

    property n:
        def __get__(self):
            return np.asarray(self._n)

        def __set__(self, val):
            self._n = val[:]

    property M1:
        def __get__(self):
            return np.asarray(self._M1)

        def __set__(self, val):
            self._M1 = val[:]

    property M2:
        def __get__(self):
            return np.asarray(self._M2)

        def __set__(self, val):
            self._M2 = val[:]



import logging
import sys

import numpy as np

from westpa.tools import WESTTool, WESTDataReader, BinMappingComponent
import westpa

from westpa.tools.binning import write_bin_info


log = logging.getLogger('w_bins')


class WBinTool(WESTTool):
    prog = 'w_bins'
    description = '''\
Display information and statistics about binning in a WEST simulation, or
modify the binning for the current iteration of a WEST simulation.
-------------------------------------------------------------------------------
'''

    def __init__(self):
        super().__init__()
        self.subcommand = None
        self.data_reader = WESTDataReader()
        self.binning = BinMappingComponent()
        self.args = None
        self.n_iter = None

    # Interface for command-line tools
    def add_args(self, parser):
        self.data_reader.add_args(parser)

        subparsers = parser.add_subparsers(help='available commands')

        info_parser = subparsers.add_parser('info', help='Display information about binning.')
        info_parser.add_argument(
            '-n', '--n-iter', type=int, help='''Consider initial points of segment N_ITER (default: current iteration).'''
        )
        info_parser.add_argument(
            '--detail',
            action='store_true',
            help='''Display detailed per-bin information in addition to summary
                                 information.''',
        )
        self.binning.add_args(info_parser)
        info_parser.set_defaults(func=self.cmd_info)

        rebin_parser = subparsers.add_parser('rebin', help='Rebuild current iteration with new binning.')
        rebin_parser.add_argument(
            '--confirm',
            action='store_true',
            help='''Commit the revised iteration to HDF5; without this option, the effects of the
                                  new binning are only calculated and printed.''',
        )
        rebin_parser.add_argument(
            '--detail',
            action='store_true',
            help='''Display detailed per-bin information in addition to summary
                                     information.''',
        )
        rebin_parser.add_argument(
            '-n', '--n-iter', type=int, help='''Consider initial points of segment N_ITER (default: current iteration).'''
        )
        self.binning.add_args(rebin_parser, suppress=['--bins-from-file'])
        self.binning.add_target_count_args(rebin_parser)
        rebin_parser.set_defaults(func=self.cmd_rebin)

    def process_args(self, args):

        self.data_reader.process_args(args)
        self.data_reader.open(mode='r+')
        self.n_iter = getattr(args, 'n_iter', None) or self.data_reader.current_iteration

        # we cannot read bin information during rebins
        # interesting note: '==' is required here; 'is' fails
        if args.func == self.cmd_rebin:
            self.binning.target_counts_required = True
        else:
            self.binning.set_we_h5file_info(self.n_iter, self.data_reader)

        self.binning.process_args(args)

        self.args = args
        self.subcommand = args.func

    def go(self):
        self.subcommand()

    def cmd_info(self):
        mapper = self.binning.mapper

        # Get target states and their assignments
        target_states = self.data_reader.get_target_states(self.n_iter)
        n_target_states = len(target_states)

        iter_group = self.data_reader.get_iter_group(self.n_iter)

        # bin initial pcoords for iteration n_iter
        initial_pcoords = iter_group['pcoord'][:, 0, :]
        assignments = mapper.assign(initial_pcoords)
        del initial_pcoords

        print('Bin information for iteration {:d}'.format(self.n_iter))

        # Get bin counts and weights
        weights = iter_group['seg_index']['weight']

        write_bin_info(mapper, assignments, weights, n_target_states, detailed=self.args.detail)

    def cmd_rebin(self):
        mapper = self.binning.mapper
        assert mapper is not None
        if self.n_iter == 1:
            sys.stderr.write('rebin is not supported for the first iteration; reinitialize with w_init instead\n')
            sys.exit(1)
        n_target_states = len(self.data_reader.get_target_states(self.n_iter))
        we_driver = westpa.rc.get_we_driver()
        data_manager = self.data_reader.data_manager

        segments = data_manager.get_segments(self.n_iter, load_pcoords=True)
        last_iter_segments = data_manager.get_segments(self.n_iter - 1, load_pcoords=False)

        # Bin on this iteration's initial points
        # We don't have to worry about recycling because we are binning on
        # initial points rather than final points, so recycling has already
        # occurred for this iteration.
        # We do need initial states, in case we merge a newly-created walker out of existence
        # avail_initial_states = {state.state_id: state
        #                        for state in data_manager.get_unused_initial_states(n_iter = self.n_iter)}
        avail_initial_states = data_manager.get_unused_initial_states(n_iter=self.n_iter)
        used_initial_states = data_manager.get_segment_initial_states(segments)
        we_driver.new_iteration(
            initial_states=avail_initial_states, bin_mapper=mapper, bin_target_counts=self.binning.bin_target_counts
        )
        we_driver.used_initial_states = {state.state_id: state for state in used_initial_states}
        we_driver.assign(segments, initializing=True)
        we_driver.rebin_current(parent_segments=last_iter_segments)

        weights = np.array([segment.weight for segment in we_driver.next_iter_segments])
        assignments = np.fromiter(we_driver.next_iter_assignments, dtype=int, count=len(weights))
        write_bin_info(mapper, assignments, weights, n_target_states, detailed=self.args.detail)

        if self.args.confirm:
            data_manager.prepare_iteration(self.n_iter, list(we_driver.next_iter_segments))

            # manually update endpoint statuses only
            endpoint_types = sorted([(segment.seg_id, segment.endpoint_type) for segment in last_iter_segments])
            last_iter_group = data_manager.get_iter_group(self.n_iter - 1)
            last_iter_index = last_iter_group['seg_index'][...]
            last_iter_index['endpoint_type'] = [pair[1] for pair in endpoint_types]
            last_iter_group['seg_index'][...] = last_iter_index

            data_manager.save_iter_binning(
                self.n_iter, self.binning.mapper_hash, self.binning.mapper_pickle, we_driver.bin_target_counts
            )
            data_manager.update_initial_states(we_driver.all_initial_states)
            data_manager.flush_backing()


def entry_point():
    WBinTool().main()


if __name__ == '__main__':
    entry_point()



import numpy as np
import scipy.optimize


def solve_steady_state(T, U, target_bins_index):
    ntarget = len(target_bins_index)  # Number of target states
    nstates = T.shape[0]  # Number of total states

    # Number of active states
    nsolve = nstates - ntarget

    # list of active states
    sactive = sorted(list(set(range(nstates)) - set(target_bins_index)))

    W = np.zeros((nsolve, nsolve))
    W_unc = np.zeros((nsolve, nsolve))
    S = np.zeros((nsolve,))

    for iiw, iit in zip(range(nsolve), sactive):
        for jjw, jjt in zip(range(nsolve), sactive):
            W[iiw, jjw] = T[jjt, iit]
            W_unc[iiw, jjw] = U[jjt, iit]

    for ii in range(nsolve):
        W[ii, ii] = 0.0
        for jj in range(nstates):  # nstates
            if jj != ii:
                W[ii, ii] -= T[ii, jj]

    # we have n equations, but only n-1 linearly independent equations
    # set one equation to constraint sum(Pi)=1

    # Pick row that contains the largest uncertainty to replace
    ii = np.unravel_index(np.argmax(W_unc), W_unc.shape)[1]

    S[ii] = 1.0
    W[ii, :] = 1.0

    # P = np.linalg.solve(W,S)
    try:
        P, err = scipy.optimize.nnls(W, S)
    except RuntimeError:
        print('Solve did not converge')
        return None

    # There are some instances where a single bin has its prob set
    # to 1.0; In this case do not reweight
    if np.allclose(np.max(P), 1.0):
        return None

    pp = np.matrix(W) * np.matrix(P.reshape(-1, 1))

    print('WESS: M*p = {}'.format(pp))
    print('WESS: max = {}'.format(np.max(pp)))

    return P, err, sactive


def prob_adjust(binprob, rates, uncert, oldindex, targets=[]):

    nbins = binprob.size

    result = solve_steady_state(rates, uncert, targets)

    if result is None:
        return binprob
    else:
        ss_estimate, err, active_bins = result

    print('WESS NNLS norm: {}'.format(err))

    # now remap new_weights onto the bin indices
    MappedActiveBins = []
    for ibin in active_bins:
        MappedActiveBins.append(oldindex[ibin])

    mapped_new_weights = np.zeros((nbins,))
    mapped_new_weights[MappedActiveBins] = ss_estimate

    # Check to make sure no bins with non-zero probability end up with zero prob after reweighting
    orig_nzi = np.nonzero(binprob)[0]  # bins that originally had non-zero prob
    new_zi = np.where(mapped_new_weights == 0.0)[0]  # bins that after reweight have zero prob
    ri = np.intersect1d(orig_nzi, new_zi)  # bins to reset
    mapped_new_weights[ri] = 1.0e-16

    # Check to make sure no bins with zero probability end up with nonzero prob after reweighting
    orig_zi = np.where(binprob == 0.0)[0]  # bins that originally had zero prob
    new_nzi = np.nonzero(mapped_new_weights)[0]  # bins that after reweight have non-zero prob
    ri = np.intersect1d(orig_zi, new_nzi)  # bins to reset
    mapped_new_weights[ri] = 0.0

    # Ensure that all target bins have their weight set to zero
    MappedTargetBins = []
    for ibin in targets:
        MappedTargetBins.append(oldindex[ibin])
    mapped_new_weights[MappedTargetBins] = 0.0

    mapped_new_weights /= np.sum(mapped_new_weights)

    assert not (mapped_new_weights < 0).any()

    return mapped_new_weights
